#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import argparse
import sys
from typing import Any

import numpy as np
from numpy.typing import NDArray
from statsmodels.robust import mad

import rapidtide.io as tide_io
import rapidtide.maskutil as tide_mask
import rapidtide.stats as tide_stats
import rapidtide.workflows.parser_funcs as pf


def _get_parser() -> Any:
    """
    Construct and return an argument parser for the atlasaverage command-line tool.

    This function builds an `argparse.ArgumentParser` object configured with all
    required and optional arguments needed to run the `atlasaverage` utility. It
    handles input validation for file paths and defines various options for
    normalizing, summarizing, and filtering data within atlas regions.

    Returns
    -------
    argparse.ArgumentParser
        Configured argument parser object for the atlasaverage tool.

    Notes
    -----
    The parser is set up with:
    - Two required positional arguments: `datafile` and `templatefile`.
    - One required positional argument: `outputroot`.
    - Several optional arguments controlling normalization, summarization,
      masking, and output formatting.

    Examples
    --------
    >>> parser = _get_parser()
    >>> args = parser.parse_args(['data.nii', 'atlas.nii', 'output'])
    >>> print(args.datafile)
    'data.nii'
    """
    # get the command line parameters
    parser = argparse.ArgumentParser(
        prog="atlasaverage",
        description="Average data within atlas regions.",
        allow_abbrev=False,
    )
    parser.add_argument(
        "datafile",
        type=lambda x: pf.is_valid_file(parser, x),
        help="The name of the 3 or 4D nifti file with the data to be averaged over atlas regions.",
    )
    parser.add_argument(
        "templatefile",
        type=lambda x: pf.is_valid_file(parser, x),
        help="The name of the atlas region NIFTI file",
    )
    parser.add_argument("outputroot", help="The root name of the output files.")

    # add optional arguments
    parser.add_argument(
        "--normmethod",
        dest="normmethod",
        action="store",
        type=str,
        choices=["none", "pct", "var", "std", "p2p"],
        help=(
            "Normalization to apply to input timecourses (in addition to demeaning) prior to "
            "combining. Choices are 'none' (no normalization, default), 'pct' (divide by mean), "
            "'var' (unit variance), 'std' (unit standard deviation), and 'p2p' (unit range)."
        ),
        default="none",
    )
    parser.add_argument(
        "--numpercentiles",
        dest="numpercentiles",
        metavar="NPCT",
        type=int,
        help=(
            "Number of evenly spaced percentiles between 0 and 100 (not including the end points) to calculate "
            "in each region.  For example, If NPCT = 1, calculate the 0th, 50th, and 100th percentiles."
        ),
        default=1,
    )
    parser.add_argument(
        "--summarymethod",
        dest="summarymethod",
        action="store",
        type=str,
        choices=["mean", "median", "sum", "std", "MAD", "CoV"],
        help=(
            "Method to summarize the voxels in a region.  Choices are 'mean' (default), 'median', 'sum', 'std', 'MAD', and 'CoV'."
        ),
        default="mean",
    )
    parser.add_argument(
        "--ignorezeros",
        dest="ignorezeros",
        action="store_true",
        help=("Zero value voxels will not be included in calculation of summary statistics."),
        default=False,
    )
    parser.add_argument(
        "--regionlistfile",
        type=lambda x: pf.is_valid_file(parser, x),
        help=(
            "The name of of a text file containing the integer region numbers to summarize, one per line.  "
            "Values that do not exist in the atlas will return NaNs."
        ),
        default=None,
    )
    parser.add_argument(
        "--regionlabelfile",
        type=lambda x: pf.is_valid_file(parser, x),
        help=(
            "The name of of a text file containing the labels of the regions, one per line.  The first line is "
            "the label integer value 1, etc."
        ),
        default=None,
    )
    parser.add_argument(
        "--includemask",
        dest="includespec",
        metavar="MASK[:VALSPEC]",
        help=(
            "Only use atlas voxels that are also in file MASK in calculating the region summaries "
            "(if VALSPEC is given, only voxels "
            "with integral values listed in VALSPEC are used). "
        ),
        default=None,
    )
    parser.add_argument(
        "--excludemask",
        dest="excludespec",
        metavar="MASK[:VALSPEC]",
        help=(
            "Do not use atlas voxels that are also in file MASK in calculating the region summaries "
            "(if VALSPEC is given, voxels "
            "with integral values listed in VALSPEC are excluded). "
        ),
        default=None,
    )
    parser.add_argument(
        "--extramask",
        dest="extramaskname",
        metavar="MASK",
        type=lambda x: pf.is_valid_file(parser, x),
        help=(
            "Additional mask to apply select voxels for region summaries. Zero voxels in this mask will be excluded."
        ),
        default=None,
    )
    parser.add_argument(
        "--headerline",
        dest="headerline",
        action="store_true",
        help="Add a header line to the text output summary of 3D files.",
        default=False,
    )
    parser.add_argument(
        "--datalabel",
        dest="datalabel",
        action="store",
        type=str,
        metavar="LABEL",
        help="Label to add to the beginning of the text summary line.",
        default=None,
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        help="Output additional debugging information.",
        default=False,
    )

    return parser


def summarizevoxels(thevoxels: NDArray, method: str = "mean") -> float:
    """
    Summarize voxel data using specified statistical method.

    Parameters
    ----------
    thevoxels : ndarray
        Input voxel data array. Can be 1D or 2D, where 2D arrays are interpreted
        as time series with shape (voxels, timepoints).
    method : str, default="mean"
        Summary method to apply. Options are:
        - "mean": Compute mean along axis 0
        - "sum": Compute sum along axis 0
        - "median": Compute median along axis 0
        - "std": Compute standard deviation along axis 0
        - "MAD": Compute median absolute deviation along axis 0
        - "CoV": Compute coefficient of variation (std/mean) along axis 0

    Returns
    -------
    float or ndarray
        Summary statistic(s) of the voxel data. Returns a scalar for 1D input
        or array of statistics for 2D input along axis 0.

    Notes
    -----
    - NaN values are converted to zero using `np.nan_to_num` before computation
    - For coefficient of variation ("CoV"), the result is multiplied by 100 to
      express as percentage
    - When input is 1D, time dimension is treated as single timepoint
    - The function handles both 1D and 2D input arrays appropriately

    Examples
    --------
    >>> import numpy as np
    >>> voxels = np.array([[1, 2, 3], [4, 5, 6]])
    >>> summarizevoxels(voxels, method="mean")
    array([2.5, 3.5, 4.5])

    >>> summarizevoxels(voxels, method="CoV")
    array([40.82482905, 33.33333333, 25.        ])
    """
    theshape = thevoxels.shape
    if len(theshape) > 1:
        numtimepoints = theshape[1]
    else:
        numtimepoints = 1

    if method == "CoV":
        if numtimepoints > 1:
            regionsummary = 100.0 * np.nan_to_num(
                np.std(thevoxels, axis=0) / np.mean(thevoxels, axis=0)
            )
        else:
            regionsummary = 100.0 * np.nan_to_num(np.std(thevoxels) / np.mean(thevoxels))
    else:
        if method == "mean":
            themethod = np.mean
        elif method == "sum":
            themethod = np.sum
        elif method == "median":
            themethod = np.median
        elif method == "std":
            themethod = np.std
        elif method == "MAD":
            themethod = mad
        else:
            print(f"illegal summary method {method} in summarizevoxels")
            sys.exit()

        if numtimepoints > 1:
            regionsummary = np.nan_to_num(themethod(thevoxels, axis=0))
        else:
            regionsummary = np.nan_to_num(themethod(thevoxels))
    return regionsummary


def atlasaverage(args: Any) -> None:
    """
    Compute average timecourses or summary statistics for regions defined by an atlas.

    This function reads fMRI data and a template (atlas) file, extracts timecourses
    or summary statistics for each region in the atlas, and saves the results to
    output files. It supports multiple normalization methods and can process both
    3D and 4D input data.

    Parameters
    ----------
    args : argparse.Namespace
        Arguments parsed from command line. Expected attributes include:
        - datafile : str
            Path to the input fMRI NIfTI file.
        - templatefile : str
            Path to the template NIfTI file defining regions.
        - normmethod : str
            Normalization method for timecourses: 'none', 'pct', 'std', 'var', 'p2p'.
        - outputroot : str
            Root name for output files.
        - debug : bool
            If True, enable debug printing and save intermediate masks.
        - includespec : str or None
            Specification for including voxels in analysis.
        - excludespec : str or None
            Specification for excluding voxels from analysis.
        - extramaskname : str or None
            Path to an additional mask file.
        - regionlabelfile : str or None
            Path to a file containing region labels.
        - regionlistfile : str or None
            Path to a file listing regions to include.
        - summarymethod : str
            Method for summarizing voxel values (e.g., 'mean', 'median').
        - datalabel : str or None
            Label to prepend to output summary.
        - ignorezeros : bool
            If True, exclude zero voxels when computing summaries.
        - numpercentiles : int
            Number of percentiles to compute for each region.
        - headerline : bool
            If True, include a header line in the summary CSV.

    Returns
    -------
    None
        This function does not return a value but writes output files to disk.

    Notes
    -----
    For 4D data, the function computes timecourses for each region and saves them
    as a TSV file. For 3D data, it computes summary statistics and saves both
    a labeled NIfTI file and a CSV/TSV summary.

    Examples
    --------
    >>> import argparse
    >>> args = argparse.Namespace(
    ...     datafile='fmri.nii.gz',
    ...     templatefile='atlas.nii.gz',
    ...     normmethod='std',
    ...     outputroot='output',
    ...     debug=False,
    ...     includespec=None,
    ...     excludespec=None,
    ...     extramaskname=None,
    ...     regionlabelfile=None,
    ...     regionlistfile=None,
    ...     summarymethod='mean',
    ...     datalabel=None,
    ...     ignorezeros=False,
    ...     numpercentiles=5,
    ...     headerline=True
    ... )
    >>> atlasaverage(args)
    """
    if args.normmethod == "none":
        print("will not normalize timecourses")
    elif args.normmethod == "pct":
        print("will normalize timecourses to percentage of mean")
    elif args.normmethod == "std":
        print("will normalize timecourses to standard deviation of 1.0")
    elif args.normmethod == "var":
        print("will normalize timecourses to variance of 1.0")
    elif args.normmethod == "p2p":
        print("will normalize timecourses to p-p deviation of 1.0")

    print("loading fmri data")
    input_img, input_data, input_hdr, thedims, thesizes = tide_io.readfromnifti(args.datafile)
    if args.debug:
        print("loading template data")
    (
        template_img,
        template_data,
        template_hdr,
        templatedims,
        templatesizes,
    ) = tide_io.readfromnifti(args.templatefile)

    print("checking dimensions")
    if not tide_io.checkspacematch(input_hdr, template_hdr):
        print("template file does not match spatial coverage of input fmri file")
        sys.exit()

    print("reshaping")
    xdim, ydim, numslices, numtimepoints = tide_io.parseniftidims(thedims)
    xsize, ysize, slicethickness, tr = tide_io.parseniftisizes(thesizes)
    numvoxels = int(xdim) * int(ydim) * int(numslices)

    templatevoxels = np.reshape(template_data, numvoxels).astype(int)
    inputvoxels = np.reshape(input_data, (numvoxels, numtimepoints))
    print("reshaped inputdata shape", inputvoxels.shape)

    # process masks
    if args.includespec is not None:
        (
            args.includename,
            args.includevals,
        ) = tide_io.processnamespec(
            args.includespec, "Including voxels where ", "in offset calculation."
        )
    else:
        args.includename = None
        args.includevals = None
    if args.excludespec is not None:
        (
            args.excludename,
            args.excludevals,
        ) = tide_io.processnamespec(
            args.excludespec, "Excluding voxels where ", "from offset calculation."
        )
    else:
        args.excludename = None
        args.excludevals = None

    includemask, excludemask, extramask = tide_mask.getmaskset(
        "anatomic",
        args.includename,
        args.includevals,
        args.excludename,
        args.excludevals,
        template_hdr,
        numvoxels,
        extramask=args.extramaskname,
    )
    themask = np.ones_like(inputvoxels[:, 0])
    if args.debug:
        print(f"{themask.shape=}")
    if includemask is not None:
        themask = themask * includemask.reshape((numvoxels))
        if args.debug:
            tide_io.savetonifti(
                includemask.reshape((xdim, ydim, numslices)),
                template_hdr,
                f"{args.outputroot}_includemask",
            )
    if excludemask is not None:
        themask = themask * (1 - excludemask.reshape((numvoxels)))
        if args.debug:
            tide_io.savetonifti(
                excludemask.reshape((xdim, ydim, numslices)),
                template_hdr,
                f"{args.outputroot}_excludemask",
            )
    if extramask is not None:
        themask = themask * extramask.reshape((numvoxels))
        if args.debug:
            tide_io.savetonifti(
                extramask.reshape((xdim, ydim, numslices)),
                template_hdr,
                f"{args.outputroot}_extramask",
            )

    # get the region names
    numregions = np.max(templatevoxels)
    if args.regionlabelfile is None:
        regionlabels = []
        numdigits = int(np.log10(numregions)) + 1
        for regnum in range(1, numregions + 1):
            regionlabels.append(f"region_{str(regnum).zfill(numdigits)}")
    else:
        regionlabels = tide_io.readlabels(args.regionlabelfile)
        if len(regionlabels) != numregions:
            print(
                "Error: number of labels in label file does not match the number of regions in the template."
            )
            sys.exit()
    if args.debug:
        print(f"Region labels: {regionlabels}")

    # decide what regions we will summarize
    if args.regionlistfile is None:
        numregions = np.max(templatevoxels)
        regionlist = range(1, numregions + 1)
    else:
        regionlist = tide_io.readvec(args.regionlistfile).astype(int)
        numregions = len(regionlist)
        newlabels = []
        for theregion in range(numregions):
            newlabels.append(regionlabels[theregion])
        regionlabels = newlabels
    if args.debug:
        print(f"Region labels to use: {regionlabels}")

    timecourses = np.zeros((numregions, numtimepoints), dtype="float")
    print(f"{numregions=}, {regionlist=}")

    if numtimepoints > 1:
        print("processing 4D input file")
        for theregion in regionlist:
            theregionvoxels = (
                inputvoxels[np.where(templatevoxels * themask == theregion)[0], :] + 0.0
            )
            print(
                "extracting",
                theregionvoxels.shape[1],
                "timepoints from region",
                theregion,
                "of",
                numregions,
            )

            # demean
            themeans = np.mean(theregionvoxels, axis=1)
            theregionvoxels -= themeans[:, None]

            if args.normmethod == "none":
                thenormfac = np.ones_like(themeans)
            elif args.normmethod == "pct":
                thenormfac = themeans
            elif args.normmethod == "var":
                thenormfac = np.var(theregionvoxels, axis=1)
            elif args.normmethod == "std":
                thenormfac = np.std(theregionvoxels, axis=1)
            elif args.normmethod == "p2p":
                thenormfac = np.max(theregionvoxels, axis=1) - np.min(theregionvoxels, axis=1)
            else:
                print("illegal normalization method", args.normmethod)
                _get_parser().print_help()
                raise
            if args.debug:
                print(theregionvoxels.shape, thenormfac.shape)
            for theloc in range(theregionvoxels.shape[0]):
                if thenormfac[theloc] != 0.0:
                    theregionvoxels[theloc, :] /= thenormfac[theloc]
            if theregionvoxels.shape[1] > 0:
                timecourses[theregion - 1, :] = summarizevoxels(
                    theregionvoxels, method=args.summarymethod
                )
        if args.debug:
            print("timecourses shape:", timecourses.shape)
        tide_io.writebidstsv(
            args.outputroot,
            timecourses,
            1.0 / tr,
            columns=regionlabels,
            yaxislabel="delay offset",
        )
    else:
        print("processing 3D input file")
        outputvoxels = np.zeros_like(inputvoxels)
        thereglabels = []
        thevals = []
        thepercentiles = []
        theregsizes = []
        thefracs = np.linspace(0.0, 1.0, args.numpercentiles + 2, endpoint=True).tolist()
        numsubregions = len(thefracs) - 1
        segmentedatlasvoxels = np.zeros_like(inputvoxels)
        if args.debug:
            print(f"{len(regionlist)=}, {regionlist=}")
            print(f"{len(regionlabels)=}, {regionlabels=}")
        if args.datalabel is not None:
            thereglabels.append("Region")
            thevals.append(args.datalabel)
        for theregion in regionlist:
            thereglabels.append(regionlabels[theregion - 1])
            theregionvoxels = inputvoxels[np.where(templatevoxels * themask == theregion)]
            initnum = theregionvoxels.shape[0]
            if args.ignorezeros:
                theregionvoxels = theregionvoxels[np.where(theregionvoxels != 0.0)]
                numremoved = theregionvoxels.shape[0] - initnum
                if numremoved > 0:
                    extrabit = f" ({numremoved} voxels removed)"
                else:
                    extrabit = ""
                if args.debug:
                    print(
                        f"extracting {theregionvoxels.shape[0]} "
                        f"non-zero voxels from region {theregion} of {numregions}{extrabit} "
                        f"({thereglabels[-1]})"
                    )
            else:
                if args.debug:
                    print(
                        f"extracting {theregionvoxels.shape[0]} "
                        f"voxels from region {theregion} of {numregions} "
                        f"({thereglabels[-1]})"
                    )
            if theregionvoxels.shape[0] > 0:
                regionval = summarizevoxels(theregionvoxels, method=args.summarymethod)
                regionsizes = theregionvoxels.shape[0]
                regionpercentiles = [
                    f"{num:.4f}"
                    for num in tide_stats.getfracvals(
                        theregionvoxels,
                        thefracs,
                        nozero=True,
                        debug=False,
                    )
                ]
                outputvoxels[np.where(templatevoxels == theregion)] = regionval
                thevals.append(str(regionval))
                theregsizes.append(str(regionsizes))
                thepercentiles.append(regionpercentiles)
            else:
                if args.debug:
                    print(f"\tregion {theregion} is empty")
                thevals.append("None")
            for thesubregion in range(numsubregions):
                scratchvoxels = np.zeros_like(inputvoxels)
                subregionkey = 1 + (theregion - 1) * numsubregions + thesubregion
                lowerlim = float(regionpercentiles[thesubregion])
                upperlim = float(regionpercentiles[thesubregion + 1])
                scratchvoxels[np.where(lowerlim <= inputvoxels)] = subregionkey
                if thesubregion < numsubregions - 1:
                    scratchvoxels[np.where(inputvoxels >= upperlim)] = 0
                scratchvoxels[np.where(templatevoxels * themask != theregion)] = 0
                segmentedatlasvoxels += scratchvoxels
        template_hdr["dim"][4] = 1
        tide_io.savetonifti(
            outputvoxels.reshape((xdim, ydim, numslices)),
            template_hdr,
            args.outputroot,
        )
        tide_io.savetonifti(
            segmentedatlasvoxels.reshape((xdim, ydim, numslices)),
            template_hdr,
            args.outputroot + "_percentiles",
        )

        if args.includename is not None or args.excludename is not None:
            tide_io.savetonifti(
                (templatevoxels * themask).reshape((xdim, ydim, numslices)),
                template_hdr,
                f"{args.outputroot}_maskedatlas",
            )
        if args.headerline:
            tide_io.writevec(
                np.array([",".join(thereglabels), ",".join(thevals)]),
                f"{args.outputroot}_regionsummaries.csv",
            )
        else:
            tide_io.writevec(
                np.array([",".join(thevals)]),
                f"{args.outputroot}_regionsummaries.csv",
            )

        outlines = []
        pctstrings = [f"{num:.0f}" for num in (np.array(thefracs) * 100.0).tolist()]
        outlines.append("Region\tVoxels\t" + "pct-" + "\tpct-".join(pctstrings))
        for idx, region in enumerate(thereglabels):
            outlines.append(
                region + "\t" + theregsizes[idx] + "\t" + "\t".join(thepercentiles[idx])
            )
            tide_io.writevec(
                np.array(outlines),
                f"{args.outputroot}_regionpercentiles.tsv",
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import argparse
import copy
import logging
import os
import sys
from argparse import Namespace
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
from numpy.typing import NDArray

import rapidtide.io as tide_io
import rapidtide.linfitfiltpass as tide_linfitfiltpass
import rapidtide.makelaggedtcs as tide_makelagged
import rapidtide.multiproc as tide_multiproc
import rapidtide.resample as tide_resample
import rapidtide.util as tide_util
import rapidtide.workflows.parser_funcs as pf

LGR = logging.getLogger("GENERAL")
ErrorLGR = logging.getLogger("ERROR")
TimingLGR = logging.getLogger("TIMING")

DEFAULT_REGRESSIONFILTDERIVS = 0


def _get_parser() -> Any:
    """
    Argument parser for retrolagtcs.

    This function constructs and returns an `argparse.ArgumentParser` object configured
    for the `retrolagtcs` command-line tool. It defines all required and optional
    arguments needed to generate voxel-specific lagged timecourses from rapidtide
    analysis maps.

    Returns
    -------
    argparse.ArgumentParser
        Configured argument parser for retrolagtcs.

    Notes
    -----
    The parser expects several input files and parameters to define the processing
    pipeline. The function uses `pf.is_valid_file` to validate the existence of
    the input 4D NIfTI file.

    Examples
    --------
    >>> parser = _get_parser()
    >>> args = parser.parse_args()
    >>> print(args.fmrifile)
    'sub-01_task-rest_bold.nii.gz'
    """
    parser = argparse.ArgumentParser(
        prog="retrolagtcs",
        description="Generate voxel specific lagged timecourses using the maps generated from a previous rapidtide analysis.",
        allow_abbrev=False,
    )

    # Required arguments
    parser.add_argument(
        "fmrifile",
        type=lambda x: pf.is_valid_file(parser, x),
        help="The name of 4D nifti fmri target file.",
    )
    parser.add_argument(
        "maskfile",
        type=str,
        help="The mask file to use (usually called XXX_desc-corrfit_mask.nii.gz)",
    )
    parser.add_argument(
        "lagtimesfile",
        type=str,
        help="The  name of the lag times file (usually called XXX_desc-maxtime_map.nii.gz)",
    )
    parser.add_argument(
        "lagtcgeneratorfile",
        type=str,
        help="The root name of the lagtc generator file (usually called XXX_desc-lagtcgenerator_timeseries)",
    )
    parser.add_argument(
        "outputroot",
        type=str,
        help="Output root.",
    )
    parser.add_argument(
        "--regressderivs",
        dest="regressderivs",
        action="store",
        type=int,
        metavar="NDERIVS",
        help=(
            f"When doing final GLM, include derivatives up to NDERIVS order. Default is {DEFAULT_REGRESSIONFILTDERIVS}"
        ),
        default=DEFAULT_REGRESSIONFILTDERIVS,
    )
    parser.add_argument(
        "--nprocs",
        dest="nprocs",
        action="store",
        type=int,
        metavar="NPROCS",
        help=(
            "Use NPROCS worker processes for multiprocessing. "
            "Setting NPROCS to less than 1 sets the number of "
            "worker processes to n_cpus."
        ),
        default=1,
    )
    parser.add_argument(
        "--numskip",
        dest="numskip",
        action="store",
        type=int,
        metavar="NUMSKIP",
        help=("Skip NUMSKIP points at the beginning of the fmri file."),
        default=0,
    )
    parser.add_argument(
        "--noprogressbar",
        dest="showprogressbar",
        action="store_false",
        help=("Will disable showing progress bars (helpful if stdout is going to a file)."),
        default=True,
    )

    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        help=("Output lots of helpful information."),
        default=False,
    )
    return parser


def retrolagtcs(args: Any) -> None:
    """
    Generate lagged time series regressors from fMRI data using a lag-time map and a generator file.

    This function reads fMRI data, a mask, and a lag-time map to compute lagged time series
    regressors for each voxel in the mask. It supports both single-process and multi-process
    execution using shared memory. The computed regressors are saved as NIfTI files.

    Parameters
    ----------
    args : Any
        An object containing the following attributes:
        - fmrifile : str
            Path to the input fMRI NIfTI file.
        - maskfile : str
            Path to the processing mask NIfTI file.
        - lagtimesfile : str
            Path to the lag times NIfTI file.
        - lagtcgeneratorfile : str
            Path to the lagtc generator file (used for resampling).
        - outputroot : str
            Root path for output files.
        - nprocs : int
            Number of processes to use for parallel execution. If less than 1, defaults to max CPU count.
        - numskip : int
            Number of initial time points to skip.
        - regressderivs : int
            Number of time derivatives to include in the regressor set.
        - showprogressbar : bool
            Whether to display a progress bar during processing.
        - debug : bool
            Whether to enable debug mode for additional logging.

    Returns
    -------
    None
        This function does not return a value but writes output files to disk.

    Notes
    -----
    - The function requires the input files to have matching spatial dimensions.
    - Shared memory is used for multi-process execution to improve performance.
    - Output files include:
        - Regressor time series (4D NIfTI)
        - Optional mask and lag time maps (3D NIfTI)

    Examples
    --------
    >>> import argparse
    >>> args = argparse.Namespace(
    ...     fmrifile='fmri.nii.gz',
    ...     maskfile='mask.nii.gz',
    ...     lagtimesfile='lagtimes.nii.gz',
    ...     lagtcgeneratorfile='generator.txt',
    ...     outputroot='output',
    ...     nprocs=4,
    ...     numskip=5,
    ...     regressderivs=2,
    ...     showprogressbar=True,
    ...     debug=False
    ... )
    >>> retrolagtcs(args)
    """
    rt_floattype = np.float64
    rt_outfloattype = np.float64

    # get the pid of the parent process
    args.pid = os.getpid()

    thecommandline = " ".join(sys.argv[1:])

    if args.nprocs < 1:
        args.nprocs = tide_multiproc.maxcpus()
    # don't use shared memory if there is only one process
    if args.nprocs == 1:
        usesharedmem = False
    else:
        usesharedmem = True

    # read the fmri input files
    print("reading fmrifile header")
    fmri_input, dummy, fmri_header, fmri_dims, fmri_sizes = tide_io.readfromnifti(
        args.fmrifile,
        headeronly=True,
    )
    xdim, ydim, slicedim, fmritr = tide_io.parseniftisizes(fmri_sizes)
    xsize, ysize, numslices, timepoints = tide_io.parseniftidims(fmri_dims)
    numspatiallocs = int(xsize) * int(ysize) * int(numslices)

    # read the processed mask
    print("reading procfit maskfile")
    (
        procmask_input,
        procmask,
        procmask_header,
        procmask_dims,
        procmask_sizes,
    ) = tide_io.readfromnifti(args.maskfile)
    if not tide_io.checkspacematch(fmri_header, procmask_header):
        raise ValueError("procmask dimensions do not match fmri dimensions")
    procmask_spacebytime = procmask.reshape((numspatiallocs))
    if args.debug:
        print(f"{procmask_spacebytime.shape=}")

    print("reading lagtimes")
    (
        lagtimes_input,
        lagtimes,
        lagtimes_header,
        lagtimes_dims,
        lagtimes_sizes,
    ) = tide_io.readfromnifti(args.lagtimesfile)
    if not tide_io.checkspacematch(fmri_header, lagtimes_header):
        raise ValueError("lagtimes dimensions do not match fmri dimensions")
    if args.debug:
        print(f"{lagtimes.shape=}")
    lagtimes_spacebytime = lagtimes.reshape((numspatiallocs))
    if args.debug:
        print(f"{lagtimes_spacebytime.shape=}")

    startpt = args.numskip
    endpt = timepoints - 1
    validtimepoints = endpt - startpt + 1
    skiptime = startpt * fmritr
    initial_fmri_x = (
        np.linspace(0.0, validtimepoints * fmritr, num=validtimepoints, endpoint=False) + skiptime
    )

    # read the lagtc generator file
    print("reading lagtc generator")
    genlagtc = tide_resample.FastResamplerFromFile(args.lagtcgeneratorfile)

    # select the voxels in the mask
    print("figuring out valid voxels")
    validvoxels = np.where(procmask_spacebytime > 0)[0]
    if args.debug:
        print(f"{validvoxels.shape=}")
    numvalidspatiallocs = np.shape(validvoxels)[0]
    if args.debug:
        print(f"{numvalidspatiallocs=}")
    internalvalidspaceshape = numvalidspatiallocs
    internalvalidspaceshapederivs = (
        internalvalidspaceshape,
        args.regressderivs + 1,
    )
    internalvalidfmrishape = (numvalidspatiallocs, np.shape(initial_fmri_x)[0])
    if args.debug:
        print(f"validvoxels shape = {numvalidspatiallocs}")
        print(f"internalvalidfmrishape shape = {internalvalidfmrishape}")

    # slicing to valid voxels
    print("selecting valid voxels")
    lagtimes_valid = lagtimes_spacebytime[validvoxels]
    procmask_valid = procmask_spacebytime[validvoxels]
    if args.debug:
        print(f"{lagtimes_valid.shape=}")

    if usesharedmem:
        if args.debug:
            print("allocating shared memory")
    else:
        if args.debug:
            print("allocating memory")
    fitNorm, fitNorm_shm = tide_util.allocarray(
        internalvalidspaceshapederivs, rt_outfloattype, shared=usesharedmem
    )
    fitcoeff, fitcoeff_shm = tide_util.allocarray(
        internalvalidspaceshapederivs, rt_outfloattype, shared=usesharedmem
    )
    lagtc, lagtc_shm = tide_util.allocarray(
        internalvalidfmrishape, rt_floattype, shared=usesharedmem
    )

    outputpath = os.path.dirname(args.outputroot)
    rawsources = [
        os.path.relpath(args.fmrifile, start=outputpath),
        os.path.relpath(args.lagtimesfile, start=outputpath),
        os.path.relpath(args.maskfile, start=outputpath),
        os.path.relpath(args.lagtcgeneratorfile, start=outputpath),
    ]

    bidsbasedict = {
        "RawSources": rawsources,
        "Units": "arbitrary",
        "CommandLineArgs": thecommandline,
    }

    print("calling makelaggedtcs")
    voxelsprocessed_makelagged = tide_makelagged.makelaggedtcs(
        genlagtc,
        initial_fmri_x,
        procmask_valid,
        lagtimes_valid,
        lagtc,
        LGR=LGR,
        nprocs=args.nprocs,
        showprogressbar=args.showprogressbar,
    )

    print(f"generated regressors for {voxelsprocessed_makelagged} voxels")

    theheader = copy.deepcopy(lagtimes_header)
    bidsdict = bidsbasedict.copy()

    if args.debug:
        maplist = [
            (
                lagtimes_valid,
                "maxtimeREAD",
                "map",
                "second",
                "Lag time in seconds used for calculation",
            ),
            (procmask_valid, "maskREAD", "mask", None, "Mask used for calculation"),
        ]

        # write the 3D maps
        tide_io.savemaplist(
            args.outputroot,
            maplist,
            validvoxels,
            (xsize, ysize, numslices),
            theheader,
            bidsdict,
            debug=args.debug,
        )

    # write the 4D maps
    theheader = copy.deepcopy(fmri_header)
    maplist = []

    if args.regressderivs > 0:
        if args.debug:
            print(f"adding derivatives up to order {args.regressderivs} prior to regression")
        regressorset = tide_linfitfiltpass.makevoxelspecificderivs(lagtc, args.regressderivs)

        if args.debug:
            print("going down the multiple EV path")
            print(f"{regressorset[:, :, 0].shape=}")
        maplist += [
            (
                regressorset[:, :, 0],
                "lfofilterEV",
                "bold",
                None,
                "Shifted sLFO regressor to filter",
            ),
        ]
        for thederiv in range(1, args.regressderivs + 1):
            if args.debug:
                print(f"{regressorset[:, :, thederiv].shape=}")
            maplist += [
                (
                    regressorset[:, :, thederiv],
                    f"lfofilterEVDeriv{thederiv}",
                    "bold",
                    None,
                    f"Time derivative {thederiv} of shifted sLFO regressor",
                ),
            ]
    else:
        regressorset = lagtc
        if args.debug:
            print("going down the single EV path")
        maplist += [
            (
                regressorset,
                "lfofilterEV",
                "bold",
                None,
                "Shifted sLFO regressor to filter",
            ),
        ]
    tide_io.savemaplist(
        args.outputroot,
        maplist,
        validvoxels,
        (xsize, ysize, numslices, validtimepoints),
        theheader,
        bidsdict,
        debug=args.debug,
    )

    # clean up shared memory
    if usesharedmem:
        tide_util.cleanup_shm(fitNorm_shm)
        tide_util.cleanup_shm(fitcoeff_shm)
        tide_util.cleanup_shm(lagtc_shm)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import sys
import warnings
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import scipy as sp
from numpy.polynomial import Polynomial
from numpy.typing import NDArray
from scipy.optimize import curve_fit
from statsmodels.robust import mad

import rapidtide.correlate as tide_corr
import rapidtide.filter as tide_filt
import rapidtide.fit as tide_fit
import rapidtide.miscmath as tide_math
import rapidtide.util as tide_util


class SimilarityFunctionator:
    reftc = None
    prepreftc = None
    testtc = None
    preptesttc = None
    timeaxis = None
    similarityfunclen = 0
    datavalid = False
    timeaxisvalid = False
    similarityfuncorigin = 0

    def __init__(
        self,
        Fs=0.0,
        similarityfuncorigin=0,
        lagmininpts=0,
        lagmaxinpts=0,
        ncprefilter=None,
        negativegradient=False,
        reftc=None,
        reftcstart=0.0,
        detrendorder=1,
        filterinputdata=True,
        debug=False,
    ):
        """
        Initialize the similarity function analysis object.

        Parameters
        ----------
        Fs : float, optional
            Sampling frequency in Hz. Default is 0.0.
        similarityfuncorigin : int, optional
            Origin point for similarity function calculation. Default is 0.
        lagmininpts : int, optional
            Minimum lag in samples. Default is 0.
        lagmaxinpts : int, optional
            Maximum lag in samples. Default is 0.
        ncprefilter : array-like, optional
            Pre-filter for cross-correlation calculation. Default is None.
        negativegradient : bool, optional
            Flag to indicate if negative gradient should be used. Default is False.
        reftc : array-like, optional
            Reference time course for cross-correlation. Default is None.
        reftcstart : float, optional
            Start time for reference time course. Default is 0.0.
        detrendorder : int, optional
            Order of detrending to apply to data. Default is 1.
        filterinputdata : bool, optional
            Flag to indicate if input data should be filtered. Default is True.
        debug : bool, optional
            Flag to enable debug mode. Default is False.

        Returns
        -------
        None
            This method initializes the object attributes but does not return any value.

        Notes
        -----
        The initialization sets up all necessary parameters for cross-correlation analysis.
        If a reference time course is provided, it is set using the setreftc method.
        All lag parameters are converted to integers by adding 0 to ensure proper type handling.

        Examples
        --------
        >>> obj = CrossCorrelationAnalyzer(Fs=100.0, lagmininpts=-10, lagmaxinpts=10)
        >>> obj = CrossCorrelationAnalyzer(reftc=reference_data, reftcstart=5.0)
        """
        self.setFs(Fs)
        self.similarityfuncorigin = similarityfuncorigin
        self.lagmininpts = lagmininpts + 0
        self.lagmaxinpts = lagmaxinpts + 0
        self.ncprefilter = ncprefilter
        self.negativegradient = negativegradient
        self.reftc = reftc
        self.detrendorder = detrendorder
        self.filterinputdata = filterinputdata
        self.debug = debug
        if self.reftc is not None:
            self.setreftc(self.reftc)
            self.reftcstart = reftcstart + 0.0

    def setFs(self, Fs: float) -> None:
        """Set the sampling frequency for the system.

        Parameters
        ----------
        Fs : float
            Sampling frequency in Hz. This parameter determines the rate at which
            samples are taken from the continuous signal.

        Returns
        -------
        None
            This method does not return any value.

        Notes
        -----
        The sampling frequency is a critical parameter that affects the
        resolution and accuracy of digital signal processing operations.
        It should be set appropriately based on the Nyquist criterion to
        avoid aliasing.

        Examples
        --------
        >>> obj = MyClass()
        >>> obj.setFs(44100.0)
        >>> print(obj.Fs)
        44100.0
        """
        self.Fs = Fs

    def preptc(self, thetc: NDArray, isreftc: bool = False) -> NDArray:
        """
        Prepare timecourse by filtering, normalizing, detrending, and applying a window function.

        This function applies a series of preprocessing steps to the input timecourse, including
        optional filtering, normalization, detrending, and window function application. The specific
        processing steps depend on the input parameters and class configuration.

        Parameters
        ----------
        thetc : NDArray
            Input timecourse data to be prepared
        isreftc : bool, optional
            Flag indicating whether the input is a reference timecourse. If True, the timecourse
            is filtered using the class's prefilter and then normalized. Default is False.

        Returns
        -------
        NDArray
            Prepared and normalized timecourse data after filtering, normalization, detrending,
            and window function application

        Notes
        -----
        The preprocessing pipeline applies the following steps in order:
        1. Filtering (if applicable based on isreftc and filterinputdata flags)
        2. Gradient calculation (when negativegradient is True)
        3. Normalization with detrending and window function application

        When isreftc is True, the input is filtered using self.ncprefilter.apply() before
        normalization. When isreftc is False and negativegradient is True, the negative gradient
        of the filtered timecourse is used. Otherwise, the filtering behavior depends on the
        filterinputdata flag.

        Examples
        --------
        >>> # Prepare a timecourse with default settings
        >>> prepared_tc = processor.preptc(input_tc)
        >>>
        >>> # Prepare a reference timecourse
        >>> ref_tc = processor.preptc(input_tc, isreftc=True)
        """
        # prepare timecourse by filtering, normalizing, detrending, and applying a window function
        if isreftc:
            thenormtc = tide_math.corrnormalize(
                self.ncprefilter.apply(self.Fs, thetc),
                detrendorder=self.detrendorder,
                windowfunc=self.windowfunc,
            )
        else:
            if self.negativegradient:
                thenormtc = tide_math.corrnormalize(
                    -np.gradient(self.ncprefilter.apply(self.Fs, thetc)),
                    detrendorder=self.detrendorder,
                    windowfunc=self.windowfunc,
                )
            else:
                if self.filterinputdata:
                    thenormtc = tide_math.corrnormalize(
                        self.ncprefilter.apply(self.Fs, thetc),
                        detrendorder=self.detrendorder,
                        windowfunc=self.windowfunc,
                    )
                else:
                    thenormtc = tide_math.corrnormalize(
                        thetc,
                        detrendorder=self.detrendorder,
                        windowfunc=self.windowfunc,
                    )

        return thenormtc

    def trim(self, vector: NDArray) -> NDArray:
        """
        Trim vector based on similarity function origin and lag constraints.

        Parameters
        ----------
        vector : NDArray
            Input vector to be trimmed.

        Returns
        -------
        NDArray
            Trimmed vector containing elements from
            `self.similarityfuncorigin - self.lagmininpts` to
            `self.similarityfuncorigin + self.lagmaxinpts`.

        Notes
        -----
        This function extracts a subset of the input vector based on the origin point
        of the similarity function and the minimum/maximum lag constraints. The trimming
        ensures that only relevant portions of the vector are considered for similarity
        calculations.

        Examples
        --------
        >>> # Assuming self.similarityfuncorigin = 10, self.lagmininpts = 2, self.lagmaxinpts = 3
        >>> trimmed_vector = trim(vector)
        >>> # Returns vector[8:13] where 8 = 10 - 2 and 13 = 10 + 3
        """
        return vector[
            self.similarityfuncorigin
            - self.lagmininpts : self.similarityfuncorigin
            + self.lagmaxinpts
        ]

    def getfunction(self, trim: bool = True) -> tuple[NDArray | None, NDArray | None, int | None]:
        """
        Retrieve simulation function data with optional trimming.

        This method returns the simulation function data and time axis, with optional
        trimming based on the trim parameter. The method handles different validation
        states of the data and returns appropriate tuples of (function, time_axis, max_value)
        or None values depending on the data validity.

        Parameters
        ----------
        trim : bool, optional
            If True, trims the simulation function and time axis using the internal
            trim method. If False, returns the raw data without trimming. Default is True.

        Returns
        -------
        tuple
            A tuple containing:
            - NDArray or None: Trimmed or untrimmed simulation function data
            - NDArray or None: Trimmed or untrimmed time axis data
            - int or None: Global maximum value, or None if not available

        Notes
        -----
        The method checks the validity of data through `self.datavalid` and `self.timeaxisvalid` attributes.
        If `self.datavalid` is True, returns both function and time axis data.
        If `self.datavalid` is False but `self.timeaxisvalid` is True, returns only time axis data.
        If neither is valid, prints an error message and returns (None, None, None).

        Examples
        --------
        >>> result = obj.getfunction(trim=True)
        >>> func_data, time_data, max_val = result
        >>>
        >>> result = obj.getfunction(trim=False)
        >>> func_data, time_data, max_val = result
        """
        if self.datavalid:
            if trim:
                return (
                    self.trim(self.thesimfunc),
                    self.trim(self.timeaxis),
                    self.theglobalmax,
                )
            else:
                return self.thesimfunc, self.timeaxis, self.theglobalmax
        else:
            if self.timeaxisvalid:
                if trim:
                    return None, self.trim(self.timeaxis), None
                else:
                    return None, self.timeaxis, None
            else:
                print("must calculate similarity function before fetching data")
                return None, None, None


class MutualInformationator(SimilarityFunctionator):
    def __init__(
        self,
        windowfunc: str = "hamming",
        norm: bool = True,
        madnorm: bool = False,
        smoothingtime: float = -1.0,
        bins: int = 20,
        sigma: float = 0.25,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize the MutualInformationator object with specified parameters.

        Parameters
        ----------
        windowfunc : str, optional
            Window function to use for spectral analysis. Default is "hamming".
        norm : bool, optional
            Whether to normalize the data. Default is True.
        madnorm : bool, optional
            Whether to use median absolute deviation normalization. Default is False.
        smoothingtime : float, optional
            Time scale for smoothing filter. If > 0, a noncausal filter is set up.
            Default is -1.0 (no smoothing).
        bins : int, optional
            Number of bins for histogram-based calculations. Default is 20.
        sigma : float, optional
            Standard deviation for Gaussian smoothing. Default is 0.25.
        *args : Any
            Additional positional arguments passed to parent class.
        **kwargs : Any
            Additional keyword arguments passed to parent class.

        Returns
        -------
        None
            This method initializes the object in-place and does not return anything.

        Notes
        -----
        When `smoothingtime` is positive, a noncausal filter is initialized with
        frequency settings based on the specified smoothing time scale.

        Examples
        --------
        >>> mi = MutualInformationator(windowfunc="hanning", bins=30, smoothingtime=2.0)
        >>> mi = MutualInformationator(norm=False, madnorm=True, sigma=0.5)
        """
        self.windowfunc = windowfunc
        self.norm = norm
        self.madnorm = madnorm
        self.bins = bins
        self.sigma = sigma
        self.smoothingtime = smoothingtime
        self.smoothingfilter = tide_filt.NoncausalFilter(filtertype="arb")
        self.mi_norm = 1.0
        if self.smoothingtime > 0.0:
            self.smoothingfilter.setfreqs(
                0.0, 0.0, 1.0 / self.smoothingtime, 1.0 / self.smoothingtime
            )
        super(MutualInformationator, self).__init__(*args, **kwargs)

    def setlimits(self, lagmininpts: int, lagmaxinpts: int) -> None:
        """
        Set the minimum and maximum lag limits for the analysis.

        This function configures the lag limits based on the provided parameters and
        adjusts the smoothing filter padding time if necessary to ensure proper
        signal processing behavior.

        Parameters
        ----------
        lagmininpts : int
            The minimum lag value in terms of number of points.
        lagmaxinpts : int
            The maximum lag value in terms of number of points.

        Returns
        -------
        None
            This function does not return any value.

        Notes
        -----
        The function automatically adjusts the smoothing filter padding time to be
        no larger than the total time span of the data. If the adjustment is made,
        a message is printed to indicate the new padding time value.

        Examples
        --------
        >>> setlimits(10, 100)
        >>> # Sets minimum lag to 10 points and maximum lag to 100 points
        """
        self.lagmininpts = lagmininpts
        self.lagmaxinpts = lagmaxinpts
        origpadtime = self.smoothingfilter.getpadtime()
        timespan = self.timeaxis[-1] - self.timeaxis[0]
        newpadtime = np.min([origpadtime, timespan])
        if newpadtime < origpadtime:
            print("lowering smoothing filter pad time to", newpadtime)
            self.smoothingfilter.setpadtime(newpadtime)

    def setbins(self, bins: int) -> None:
        """
        Set the number of bins for histogram calculation.

        Parameters
        ----------
        bins : int
            The number of bins to use for histogram calculation.

        Returns
        -------
        None
            This method does not return any value.

        Notes
        -----
        This method assigns the specified number of bins to the instance variable
        `self.bins`. The bins parameter determines the granularity of the histogram
        distribution.

        Examples
        --------
        >>> obj = MyClass()
        >>> obj.setbins(10)
        >>> print(obj.bins)
        10
        """
        self.bins = bins

    def setreftc(self, reftc: NDArray, offset: float = 0.0) -> None:
        """
        Set reference time course and compute cross-mutual information.

        This method initializes the reference time course and computes the cross-mutual
        information between the reference time course and itself to determine the
        optimal time alignment and similarity function.

        Parameters
        ----------
        reftc : NDArray
            Reference time course array to be set and processed.
        offset : float, optional
            Time offset to be applied to the time axis (default is 0.0).

        Returns
        -------
        None
            This method modifies the object's attributes in-place and does not return anything.

        Notes
        -----
        The method performs the following operations:
        1. Stores a copy of the reference time course
        2. Pre-processes the reference time course using preptc method
        3. Computes cross-mutual information using tide_corr.cross_mutual_info
        4. Adjusts the time axis by the specified offset
        5. Sets various internal attributes including similarity function normalization

        Examples
        --------
        >>> obj.setreftc(reference_data, offset=0.5)
        >>> print(obj.timeaxis)
        >>> print(obj.similarityfunclen)
        """
        self.reftc = reftc + 0.0
        self.prepreftc = self.preptc(self.reftc, isreftc=True)

        self.timeaxis, self.automi, self.similarityfuncorigin = tide_corr.cross_mutual_info(
            self.prepreftc,
            self.prepreftc,
            Fs=self.Fs,
            fast=True,
            negsteps=self.lagmininpts,
            possteps=self.lagmaxinpts,
            returnaxis=True,
        )

        self.timeaxis -= offset
        self.similarityfunclen = len(self.timeaxis)
        self.timeaxisvalid = True
        self.datavalid = False
        self.mi_norm = np.nan_to_num(1.0 / np.max(self.automi))
        if self.debug:
            print(f"MutualInformationator setreftc: {len(self.timeaxis)=}")
            print(f"MutualInformationator setreftc: {self.timeaxis}")
            print(f"MutualInformationator setreftc: {self.mi_norm=}")

    def getnormfac(self) -> float:
        """
        Return the normalization factor stored in the instance.

        This method provides access to the normalization factor that has been
        previously computed and stored in the instance variable `mi_norm`.

        Returns
        -------
        float
            The normalization factor value stored in `self.mi_norm`.

        Notes
        -----
        The normalization factor is typically used to scale or normalize
        data within the class. This value should be set before calling this
        method to ensure meaningful results.

        Examples
        --------
        >>> instance = MyClass()
        >>> instance.mi_norm = 2.5
        >>> norm_factor = instance.getnormfac()
        >>> print(norm_factor)
        2.5
        """
        return self.mi_norm

    def run(
        self,
        thetc: NDArray,
        locs: NDArray | None = None,
        trim: bool = True,
        gettimeaxis: bool = True,
    ) -> tuple[NDArray, NDArray, int] | NDArray:
        """
        Compute cross-mutual information between test and reference timecourses.

        This function calculates the cross-mutual information between a test timecourse
        and a reference timecourse, optionally applying preprocessing, trimming, and
        smoothing. It supports both trimmed and untrimmed outputs, and can return
        time axis information depending on the input parameters.

        Parameters
        ----------
        thetc : NDArray
            Test timecourse array of shape (n_times,).
        locs : NDArray | None, optional
            Locations to compute mutual information at; if provided, the function
            will return only the similarity function values at these locations.
            Default is None.
        trim : bool, optional
            If True, trim the output similarity function and time axis to the
            valid range defined by `lagmininpts` and `lagmaxinpts`. If False,
            the full similarity function is returned. Default is True.
        gettimeaxis : bool, optional
            If True, return the time axis along with the similarity function.
            If False, only the similarity function is returned. Default is True.

        Returns
        -------
        tuple[NDArray, NDArray, int] | NDArray
            If `locs` is not None, returns the similarity function values at the
            specified locations.
            If `trim` is True and `gettimeaxis` is True, returns a tuple of:
            - trimmed similarity function (NDArray)
            - trimmed time axis (NDArray)
            - index of the global maximum (int)
            If `trim` is False and `gettimeaxis` is True, returns a tuple of:
            - full similarity function (NDArray)
            - full time axis (NDArray)
            - index of the global maximum (int)
            If `gettimeaxis` is False, returns only the similarity function (NDArray).

        Notes
        -----
        This function uses `tide_corr.cross_mutual_info` for computing cross-mutual
        information, and applies normalization and optional smoothing based on
        instance attributes.

        Examples
        --------
        >>> result = obj.run(test_tc, locs=None, trim=True, gettimeaxis=True)
        >>> sim_func, time_axis, max_idx = result
        """
        if len(thetc) != len(self.reftc):
            print(
                "MutualInformationator: timecourses are of different sizes:",
                len(thetc),
                "!=",
                len(self.reftc),
                "- exiting",
            )
            sys.exit()

        self.testtc = thetc
        self.preptesttc = self.preptc(self.testtc)

        if locs is not None:
            gettimeaxis = True

        # now calculate the similarity function
        if trim:
            retvals = tide_corr.cross_mutual_info(
                self.preptesttc,
                self.prepreftc,
                norm=self.norm,
                negsteps=self.lagmininpts,
                possteps=self.lagmaxinpts,
                locs=locs,
                madnorm=self.madnorm,
                returnaxis=gettimeaxis,
                fast=True,
                Fs=self.Fs,
                sigma=self.sigma,
                bins=self.bins,
            )
        else:
            retvals = tide_corr.cross_mutual_info(
                self.preptesttc,
                self.prepreftc,
                norm=self.norm,
                negsteps=-1,
                possteps=-1,
                locs=locs,
                madnorm=self.madnorm,
                returnaxis=gettimeaxis,
                fast=True,
                Fs=self.Fs,
                sigma=self.sigma,
                bins=self.bins,
            )
        if gettimeaxis:
            self.timeaxis, self.thesimfunc, self.similarityfuncorigin = (
                retvals[0],
                retvals[1],
                retvals[2],
            )
            self.timeaxisvalid = True
        else:
            self.thesimfunc = retvals[0]

        # normalize
        self.thesimfunc *= self.mi_norm

        if locs is not None:
            return self.thesimfunc

        if self.smoothingtime > 0.0:
            self.thesimfunc = self.smoothingfilter.apply(self.Fs, self.thesimfunc)

        self.similarityfunclen = len(self.thesimfunc)
        if trim:
            self.similarityfuncorigin = self.lagmininpts + 1
        else:
            self.similarityfuncorigin = self.similarityfunclen // 2 + 1

        # find the global maximum value
        self.theglobalmax = np.argmax(self.thesimfunc)
        self.datavalid = True

        # make a dummy filtered baseline
        self.filteredbaseline = np.zeros_like(self.thesimfunc)

        if trim:
            return (
                self.trim(self.thesimfunc),
                self.trim(self.timeaxis),
                self.theglobalmax,
            )
        else:
            return self.thesimfunc, self.timeaxis, self.theglobalmax


class Correlator(SimilarityFunctionator):
    def __init__(
        self,
        windowfunc: str = "hamming",
        corrweighting: str = "None",
        corrpadding: int = 0,
        baselinefilter: Any | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize the Correlator with specified parameters.

        Parameters
        ----------
        windowfunc : str, default="hamming"
            Window function to apply during correlation. Common options include
            'hamming', 'hanning', 'blackman', etc.
        corrweighting : str, default="None"
            Correlation weighting method. Can be 'None' or other weighting schemes
            depending on implementation.
        corrpadding : int, default=0
            Padding size to apply during correlation operations.
        baselinefilter : Any | None, default=None
            Baseline filtering method or object to apply. Can be None to skip filtering.
        *args : Any
            Additional positional arguments passed to parent class.
        **kwargs : Any
            Additional keyword arguments passed to parent class.

        Returns
        -------
        None
            This method initializes the instance and does not return any value.

        Notes
        -----
        The Correlator class inherits from a parent class, and this initialization
        method sets up the correlation parameters before calling the parent's
        initialization method.

        Examples
        --------
        >>> correlator = Correlator(windowfunc="hanning", corrpadding=10)
        >>> correlator = Correlator(baselinefilter=my_filter_object)
        """
        self.windowfunc = windowfunc
        self.corrweighting = corrweighting
        self.corrpadding = corrpadding
        self.baselinefilter = baselinefilter
        super(Correlator, self).__init__(*args, **kwargs)

    def setlimits(self, lagmininpts: int, lagmaxinpts: int) -> None:
        """
        Set the minimum and maximum lag limits for the analysis.

        Parameters
        ----------
        lagmininpts : int
            The minimum lag value in points for the analysis.
        lagmaxinpts : int
            The maximum lag value in points for the analysis.

        Returns
        -------
        None
            This method does not return any value.

        Notes
        -----
        This method assigns the provided lag limits to instance variables
        `self.lagmininpts` and `self.lagmaxinpts`. The lag limits define the
        range of lags to be considered in the subsequent analysis operations.

        Examples
        --------
        >>> obj = MyClass()
        >>> obj.setlimits(5, 20)
        >>> print(obj.lagmininpts)
        5
        >>> print(obj.lagmaxinpts)
        20
        """
        self.lagmininpts = lagmininpts
        self.lagmaxinpts = lagmaxinpts

    def setreftc(self, reftc: NDArray, offset: float = 0.0) -> None:
        """
        Set reference time course and initialize related attributes.

        This function sets the reference time course, computes related parameters,
        and initializes the time axis for similarity function calculations.

        Parameters
        ----------
        reftc : NDArray
            Reference time course array used for similarity calculations.
        offset : float, optional
            Time offset to apply to the reference time axis (default is 0.0).

        Returns
        -------
        None
            This function modifies instance attributes in-place and does not return anything.

        Notes
        -----
        This function performs the following operations:
        1. Creates a copy of the reference time course
        2. Computes preprocessed reference time course using preptc method
        3. Calculates similarity function length and origin
        4. Constructs time axis based on sampling frequency and offset

        The time axis is centered around zero with the specified offset applied.

        Examples
        --------
        >>> setreftc(reftc_array, offset=0.5)
        >>> print(self.timeaxis)
        >>> print(self.similarityfunclen)
        """
        self.reftc = reftc + 0.0
        self.prepreftc = self.preptc(self.reftc, isreftc=True)
        self.similarityfunclen = len(self.reftc) * 2 - 1
        self.similarityfuncorigin = self.similarityfunclen // 2 + 1

        # make the reference time axis
        self.timeaxis = (
            np.arange(0.0, self.similarityfunclen) * (1.0 / self.Fs)
            - ((self.similarityfunclen - 1) * (1.0 / self.Fs)) / 2.0
        ) - offset
        self.timeaxisvalid = True
        self.datavalid = False

    def run(self, thetc: NDArray, trim: bool = True) -> tuple[NDArray, NDArray, int]:
        """
        Compute the correlation between test and reference timecourses.

        This function performs correlation analysis between a test timecourse and a reference
        timecourse, applying preprocessing and optional filtering. It returns the similarity
        function, time axis, and the index of the global maximum.

        Parameters
        ----------
        thetc : ndarray
            Test timecourse to be correlated with the reference timecourse.
        trim : bool, optional
            If True, trims the similarity function and time axis to remove zero-padding
            effects. Default is True.

        Returns
        -------
        tuple of (ndarray, ndarray, int)
            A tuple containing:
                - similarity function (ndarray)
                - time axis (ndarray)
                - index of the global maximum (int)

        Notes
        -----
        The function exits with status code 1 if the lengths of `thetc` and `self.reftc`
        do not match.

        Examples
        --------
        >>> result = correlator.run(test_timecourse, trim=True)
        >>> similarity_func, time_axis, max_index = result
        """
        if len(thetc) != len(self.reftc):
            print(
                "Correlator: timecourses are of different sizes:",
                len(thetc),
                "!=",
                len(self.reftc),
                "- exiting",
            )
            sys.exit()

        self.testtc = thetc
        self.preptesttc = self.preptc(self.testtc)

        # now actually do the correlation
        self.thesimfunc = tide_corr.fastcorrelate(
            self.preptesttc,
            self.prepreftc,
            usefft=True,
            weighting=self.corrweighting,
            zeropadding=self.corrpadding,
            debug=self.debug,
        )
        self.similarityfunclen = len(self.thesimfunc)
        self.similarityfuncorigin = self.similarityfunclen // 2 + 1

        if self.baselinefilter is not None:
            self.filteredbaseline = self.baselinefilter.apply(self.Fs, self.thesimfunc)
        else:
            self.filteredbaseline = np.zeros_like(self.thesimfunc)

        # find the global maximum value
        self.theglobalmax = np.argmax(self.thesimfunc)
        self.datavalid = True

        if trim:
            return (
                self.trim(self.thesimfunc),
                self.trim(self.timeaxis),
                self.theglobalmax,
            )
        else:
            return self.thesimfunc, self.timeaxis, self.theglobalmax


class SimilarityFunctionFitter:
    corrtimeaxis = None
    FML_NOERROR = np.uint32(0x0000)

    FML_INITAMPLOW = np.uint32(0x0001)
    FML_INITAMPHIGH = np.uint32(0x0002)
    FML_INITWIDTHLOW = np.uint32(0x0004)
    FML_INITWIDTHHIGH = np.uint32(0x0008)
    FML_INITLAGLOW = np.uint32(0x0010)
    FML_INITLAGHIGH = np.uint32(0x0020)
    FML_INITFAIL = (
        FML_INITAMPLOW
        | FML_INITAMPHIGH
        | FML_INITWIDTHLOW
        | FML_INITWIDTHHIGH
        | FML_INITLAGLOW
        | FML_INITLAGHIGH
    )

    FML_FITAMPLOW = np.uint32(0x0100)
    FML_FITAMPHIGH = np.uint32(0x0200)
    FML_FITWIDTHLOW = np.uint32(0x0400)
    FML_FITWIDTHHIGH = np.uint32(0x0800)
    FML_FITLAGLOW = np.uint32(0x1000)
    FML_FITLAGHIGH = np.uint32(0x2000)
    FML_FITALGOFAIL = np.uint32(0x0400)
    FML_FITFAIL = (
        FML_FITAMPLOW
        | FML_FITAMPHIGH
        | FML_FITWIDTHLOW
        | FML_FITWIDTHHIGH
        | FML_FITLAGLOW
        | FML_FITLAGHIGH
        | FML_FITALGOFAIL
    )

    def __init__(
        self,
        corrtimeaxis=None,
        lagmin=-30.0,
        lagmax=30.0,
        absmaxsigma=1000.0,
        absminsigma=0.25,
        hardlimit=True,
        bipolar=False,
        lthreshval=0.0,
        uthreshval=1.0,
        debug=False,
        zerooutbadfit=True,
        maxguess=0.0,
        useguess=False,
        searchfrac=0.5,
        lagmod=1000.0,
        enforcethresh=True,
        displayplots=False,
        corrtolerance=1e-2,
        functype="correlation",
        peakfittype="gauss",
    ):
        """
        Initialize a correlation peak finder.

        This constructor sets up the parameters for fitting and searching correlation
        functions to find peak locations, amplitudes, and widths.

        Parameters
        ----------
        corrtimeaxis : 1D float array, optional
            The time axis of the correlation function. Default is None.
        lagmin : float, optional
            The minimum allowed lag time in seconds. Default is -30.0.
        lagmax : float, optional
            The maximum allowed lag time in seconds. Default is 30.0.
        absmaxsigma : float, optional
            The maximum allowed peak halfwidth in seconds. Default is 1000.0.
        absminsigma : float, optional
            The minimum allowed peak halfwidth in seconds. Default is 0.25.
        hardlimit : bool, optional
            If True, enforce hard limits on peak fitting. Default is True.
        bipolar : bool, optional
            If True, find the correlation peak with the maximum absolute value,
            regardless of sign. Default is False.
        lthreshval : float, optional
            Lower threshold value for correlation function. Default is 0.0.
        uthreshval : float, optional
            Upper threshold value for correlation function. Default is 1.0.
        debug : bool, optional
            If True, enable debug output. Default is False.
        zerooutbadfit : bool, optional
            If True, set bad fits to zero. Default is True.
        maxguess : float, optional
            Maximum guess for peak fitting. Default is 0.0.
        useguess : bool, optional
            If True, use initial guess for peak fitting. Default is False.
        searchfrac : float, optional
            Fraction of the search range to consider for peak fitting. Default is 0.5.
        lagmod : float, optional
            Modulus for lag values. Default is 1000.0.
        enforcethresh : bool, optional
            If True, enforce threshold constraints. Default is True.
        corrtolerance : float, optional
            Amount by which the magnitude of a correlation can exceed 1.0 (in a case with
            almost no noise) and not be flagged as an error. Default is 1e-6.
        displayplots : bool, optional
            If True, display plots during fitting. Default is False.
        functype : str, optional
            Type of function to fit. Either "correlation" or "mutualinfo". Default is "correlation".
        peakfittype : str, optional
            Type of peak fit to use. Default is "gauss".

        Returns
        -------
        None
            This method initializes the object and does not return any value.

        Notes
        -----
        The `corrtimeaxis` must be provided before calling `fit()` method.
        The `functype` parameter determines whether to fit a correlation or mutual information function.

        Examples
        --------
        >>> peakfinder = PeakFinder(corrtimeaxis=time_axis, lagmin=-20, lagmax=20)
        >>> peak_location, peak_value, peak_width = peakfinder.fit(correlation_data)
        """
        self.setcorrtimeaxis(corrtimeaxis)
        self.lagmin = lagmin + 0.0
        self.lagmax = lagmax + 0.0
        self.absmaxsigma = absmaxsigma + 0.0
        self.absminsigma = absminsigma + 0.0
        self.hardlimit = hardlimit
        self.bipolar = bipolar
        self.lthreshval = lthreshval + 0.0
        self.uthreshval = uthreshval + 0.0
        self.debug = debug
        if functype == "correlation" or functype == "mutualinfo":
            self.functype = functype
        else:
            print("illegal functype")
            sys.exit()
        self.peakfittype = peakfittype
        self.zerooutbadfit = zerooutbadfit
        self.maxguess = maxguess + 0.0
        self.useguess = useguess
        self.searchfrac = searchfrac + 0.0
        self.lagmod = lagmod + 0.0
        self.enforcethresh = enforcethresh
        self.corrtolerance = corrtolerance
        self.displayplots = displayplots

    def _maxindex_noedge(self, corrfunc: NDArray) -> tuple[int, float]:
        """
        Find the index of the maximum value in correlation function, avoiding edge effects.

        This function searches for the maximum value in the correlation function while
        avoiding the edges of the data. It handles bipolar correlation functions by
        considering both positive and negative peaks, returning the one with the larger
        absolute value. The function also accounts for edge effects by adjusting the
        search boundaries when the maximum is found at the edge.

        Parameters
        ----------
        corrfunc : NDArray
            Correlation function array to search for maximum value

        Returns
        -------
        tuple[int, float]
            Tuple containing:
            - maxindex: Index of the maximum value in the correlation function
            - flipfac: Flipping factor (-1.0 if minimum was selected, 1.0 otherwise)

        Notes
        -----
        The function adjusts search boundaries to avoid edge effects:
        - If maximum is at index 0, lowerlim is incremented
        - If maximum is at upper limit, upperlim is decremented
        - For bipolar correlation functions, both positive and negative peaks are considered
        - The search continues until no edge effects are detected

        Examples
        --------
        >>> max_index, flip_factor = obj._maxindex_noedge(corrfunc)
        >>> print(f"Maximum at index {max_index} with flip factor {flip_factor}")
        """
        lowerlim = 0
        upperlim = len(self.corrtimeaxis) - 1
        done = False
        while not done:
            flipfac = 1.0
            done = True
            maxindex = (np.argmax(corrfunc[lowerlim:upperlim]) + lowerlim).astype("int32")
            if self.bipolar:
                minindex = (np.argmax(-corrfunc[lowerlim:upperlim]) + lowerlim).astype("int32")
                if np.fabs(corrfunc[minindex]) > np.fabs(corrfunc[maxindex]):
                    maxindex = minindex
                    flipfac = -1.0
            if upperlim == lowerlim:
                done = True
            if maxindex == 0:
                lowerlim += 1
                done = False
            if maxindex == upperlim:
                upperlim -= 1
                done = False
        return maxindex, flipfac

    def setfunctype(self, functype: str) -> None:
        """
        Set the function type for the object.

        Parameters
        ----------
        functype : str
            The function type to be set. This should be a string identifier
            that defines the type of function this object represents.

        Returns
        -------
        None
            This method does not return any value.

        Notes
        -----
        This method directly assigns the provided function type to the
        internal `functype` attribute of the object.

        Examples
        --------
        >>> obj = MyClass()
        >>> obj.setfunctype('linear')
        >>> obj.functype
        'linear'
        """
        self.functype = functype

    def setpeakfittype(self, peakfittype: str) -> None:
        """
        Set the peak fitting type for the analysis.

        Parameters
        ----------
        peakfittype : str
            The type of peak fitting to be used. This parameter determines the
            mathematical model and fitting algorithm applied to the peak data.

        Returns
        -------
        None
            This method does not return any value.

        Notes
        -----
        This method directly assigns the provided peak fitting type to the
        instance variable `self.peakfittype`. The valid values for peakfittype
        depend on the specific implementation of the peak fitting algorithms
        available in the class.

        Examples
        --------
        >>> analyzer = PeakAnalyzer()
        >>> analyzer.setpeakfittype('gaussian')
        >>> print(analyzer.peakfittype)
        'gaussian'
        """
        self.peakfittype = peakfittype

    def setrange(self, lagmin: float, lagmax: float) -> None:
        """
        Set the range of lags for the analysis.

        Parameters
        ----------
        lagmin : float
            The minimum lag value for the analysis range.
        lagmax : float
            The maximum lag value for the analysis range.

        Returns
        -------
        None
            This method does not return any value.

        Notes
        -----
        This method updates the internal lag range parameters of the object.
        The lagmin value should be less than or equal to the lagmax value.

        Examples
        --------
        >>> obj = MyClass()
        >>> obj.setrange(0.0, 10.0)
        >>> print(obj.lagmin)
        0.0
        >>> print(obj.lagmax)
        10.0
        """
        self.lagmin = lagmin
        self.lagmax = lagmax

    def setcorrtimeaxis(self, corrtimeaxis: NDArray | None) -> None:
        """
        Set the correlation time axis for the object.

        This method assigns the provided correlation time axis to the object's
        `corrtimeaxis` attribute. If the input is not None, a copy of the array is
        created to avoid modifying the original data.

        Parameters
        ----------
        corrtimeaxis : NDArray | None
            The correlation time axis array to be set. If None, the attribute will
            be set to None. If an array is provided, a copy will be created to
            prevent modification of the original array.

        Returns
        -------
        None
            This method does not return any value.

        Notes
        -----
        When a numpy array is passed, the method creates a copy using `+ 0.0`
        to ensure that modifications to the original array do not affect the
        object's internal state.

        Examples
        --------
        >>> obj.setcorrtimeaxis(np.array([1, 2, 3, 4]))
        >>> obj.setcorrtimeaxis(None)
        """
        if corrtimeaxis is not None:
            self.corrtimeaxis = corrtimeaxis + 0.0
        else:
            self.corrtimeaxis = corrtimeaxis

    def setguess(self, useguess: bool, maxguess: float = 0.0) -> None:
        """
        Set the guess parameters for the optimization process.

        This method configures whether to use a guess value and sets the maximum
        guess value for optimization algorithms.

        Parameters
        ----------
        useguess : bool
            Flag indicating whether to use a guess value in the optimization process.
            If True, the algorithm will attempt to use the provided guess value.
            If False, no guess value will be used.
        maxguess : float, optional
            Maximum guess value to be used in the optimization process. Default is 0.0.
            This parameter is only relevant when useguess is True.

        Returns
        -------
        None
            This method does not return any value.

        Notes
        -----
        The maxguess parameter is typically used to constrain the search space
        during optimization. When useguess is False, the maxguess parameter has
        no effect on the optimization process.

        Examples
        --------
        >>> optimizer = Optimizer()
        >>> optimizer.setguess(True, 10.0)
        >>> optimizer.setguess(False)
        """
        self.useguess = useguess
        self.maxguess = maxguess

    def setlthresh(self, lthreshval: float) -> None:
        """
        Set the lower threshold value for the object.

        Parameters
        ----------
        lthreshval : float
            The lower threshold value to be set. This value will be assigned to
            the instance attribute `lthreshval`.

        Returns
        -------
        None
            This method does not return any value.

        Notes
        -----
        This method assigns the provided threshold value to the instance attribute
        `lthreshval`. The threshold value is typically used for filtering or
        processing operations where values below this threshold are treated
        differently.

        Examples
        --------
        >>> obj = MyClass()
        >>> obj.setlthresh(0.5)
        >>> print(obj.lthreshval)
        0.5
        """
        self.lthreshval = lthreshval

    def setuthresh(self, uthreshval: float) -> None:
        """
        Set the upper threshold value for the object.

        Parameters
        ----------
        uthreshval : float
            The upper threshold value to be set. This value will be assigned to
            the object's internal `uthreshval` attribute.

        Returns
        -------
        None
            This method does not return any value.

        Notes
        -----
        This method directly assigns the provided threshold value to the object's
        internal attribute. No validation or processing is performed on the input value.

        Examples
        --------
        >>> obj = MyClass()
        >>> obj.setuthresh(0.5)
        >>> print(obj.uthreshval)
        0.5
        """
        self.uthreshval = uthreshval

    def diagnosefail(self, failreason: Any) -> str:
        """
        Diagnose the cause of a failure based on bitwise flags.

        This function takes a failure reason encoded as a bitwise flag and returns
        a human-readable string describing the cause(s) of the failure. Each flag
        corresponds to a specific condition that may have led to the failure.

        Parameters
        ----------
        failreason : Any
            A value representing the failure reason, typically an integer or array
            of integers. It is cast to `np.uint32` for bitwise operations.

        Returns
        -------
        str
            A comma-separated string listing the reasons for the failure. If no
            reasons are found, returns "No error".

        Notes
        -----
        The function checks the following flags:
        - ``FML_INITAMPLOW``, ``FML_INITAMPHIGH``, ``FML_INITWIDTHLOW``,
          ``FML_INITWIDTHHIGH``, ``FML_INITLAGLOW``, ``FML_INITLAGHIGH``:
          Initial parameter values are out of bounds.
        - ``FML_FITAMPLOW``, ``FML_FITAMPHIGH``, ``FML_FITWIDTHLOW``,
          ``FML_FITWIDTHHIGH``, ``FML_FITLAGLOW``, ``FML_FITLAGHIGH``:
          Fit parameter values are out of bounds.
        - ``FML_FITALGOFAIL``: Nonlinear fitting algorithm failed.

        Examples
        --------
        >>> diagnosis = obj.diagnosefail(0x0001)
        >>> print(diagnosis)
        'Initial amplitude too low'
        """
        # define error values
        reasons = []
        if failreason.astype(np.uint32) & self.FML_INITAMPLOW:
            reasons.append("Initial amplitude too low")
        if failreason.astype(np.uint32) & self.FML_INITAMPHIGH:
            reasons.append("Initial amplitude too high")
        if failreason.astype(np.uint32) & self.FML_INITWIDTHLOW:
            reasons.append("Initial width too low")
        if failreason.astype(np.uint32) & self.FML_INITWIDTHHIGH:
            reasons.append("Initial width too high")
        if failreason.astype(np.uint32) & self.FML_INITLAGLOW:
            reasons.append("Initial Lag too low")
        if failreason.astype(np.uint32) & self.FML_INITLAGHIGH:
            reasons.append("Initial Lag too high")

        if failreason.astype(np.uint32) & self.FML_FITAMPLOW:
            reasons.append("Fit amplitude too low")
        if failreason.astype(np.uint32) & self.FML_FITAMPHIGH:
            reasons.append("Fit amplitude too high")
        if failreason.astype(np.uint32) & self.FML_FITWIDTHLOW:
            reasons.append("Fit width too low")
        if failreason.astype(np.uint32) & self.FML_FITWIDTHHIGH:
            reasons.append("Fit width too high")
        if failreason.astype(np.uint32) & self.FML_FITLAGLOW:
            reasons.append("Fit Lag too low")
        if failreason.astype(np.uint32) & self.FML_FITLAGHIGH:
            reasons.append("Fit Lag too high")
        if failreason.astype(np.uint32) & self.FML_FITALGOFAIL:
            reasons.append("Nonlinear fit failed")

        if len(reasons) > 0:
            return ", ".join(reasons)
        else:
            return "No error"

    def fit(self, incorrfunc: NDArray) -> tuple[int, float, float, float, int, Any, int, int]:
        """
        Fit a correlation function to determine peak parameters including lag, amplitude, and width.

        This function performs a fit on the provided correlation function to extract key parameters
        such as the peak lag, amplitude, and width. It supports multiple fitting methods and handles
        various edge cases including invalid inputs, out-of-bounds values, and fitting failures.

        Parameters
        ----------
        incorrfunc : ndarray
            The input correlation function to be fitted. Must match the length of `self.corrtimeaxis`.

        Returns
        -------
        tuple[int, float, float, float, int, Any, int, int]
            A tuple containing:

            - `maxindex` (int): Index of the maximum value in the correlation function.
            - `maxlag` (float): The lag corresponding to the peak, in seconds.
            - `maxval` (float): The amplitude of the peak, adjusted for flip factor.
            - `maxsigma` (float): The width of the peak, in seconds.
            - `maskval` (int): A flag indicating fit success (1) or failure (0).
            - `failreason` (Any): A bitmask indicating the reason for fit failure, if any.
            - `peakstart` (int): Start index of the peak region used in fitting.
            - `peakend` (int): End index of the peak region used in fitting.

        Notes
        -----
        The function performs several checks:

        - Ensures `self.corrtimeaxis` is defined and matches the input length.
        - Handles bipolar correlation functions and adjusts signs accordingly.
        - Applies initial parameter estimation based on the input data.
        - Supports multiple fitting algorithms including Gaussian, quadratic, and center-of-mass.
        - Applies bounds checking for lag, amplitude, and width to ensure physical validity.
        - Outputs debugging information if `self.debug` is set to True.

        Examples
        --------
        >>> # Assuming `fit_instance` is an instance of the class containing this method
        >>> corr_func = np.array([0.1, 0.5, 1.0, 0.5, 0.1])
        >>> result = fit_instance.fit(corr_func)
        >>> print(result)
        (2, 1.0, 1.0, 0.5, 1, 0, 1, 3)
        """
        # check to make sure xcorr_x and xcorr_y match
        if self.corrtimeaxis is None:
            print("Correlation time axis is not defined - exiting")
            sys.exit()
        if len(self.corrtimeaxis) != len(incorrfunc):
            print(
                "Correlation time axis and values do not match in length (",
                len(self.corrtimeaxis),
                "!=",
                len(incorrfunc),
                "- exiting",
            )
            sys.exit()
        # set initial parameters
        # absmaxsigma is in seconds
        # maxsigma is in Hz
        # maxlag is in seconds
        warnings.filterwarnings("ignore", "Number*")
        failreason = self.FML_NOERROR
        maskval = np.uint16(1)  # start out assuming the fit will succeed
        binwidth = self.corrtimeaxis[1] - self.corrtimeaxis[0]

        # set the search range
        lowerlim = 0
        upperlim = len(self.corrtimeaxis) - 1
        if self.debug:
            print(
                "initial search indices are",
                lowerlim,
                "to",
                upperlim,
                "(",
                self.corrtimeaxis[lowerlim],
                self.corrtimeaxis[upperlim],
                ")",
            )

        # make an initial guess at the fit parameters for the gaussian
        # start with finding the maximum value and its location
        flipfac = 1.0
        corrfunc = incorrfunc + 0.0
        if self.useguess:
            maxindex = tide_util.valtoindex(self.corrtimeaxis, self.maxguess)
            if (corrfunc[maxindex] < 0.0) and self.bipolar:
                flipfac = -1.0
        else:
            maxindex, flipfac = self._maxindex_noedge(corrfunc)
        corrfunc *= flipfac
        maxlag_init = (1.0 * self.corrtimeaxis[maxindex]).astype("float64")
        maxval_init = corrfunc[maxindex].astype("float64")
        if self.debug:
            print(
                "maxindex, maxlag_init, maxval_init:",
                maxindex,
                maxlag_init,
                maxval_init,
            )

        # set the baseline and baselinedev levels
        if (self.functype == "correlation") or (self.functype == "hybrid"):
            baseline = 0.0
            baselinedev = 0.0
        else:
            # for mutual information, there is a nonzero baseline, so we want the difference from that.
            baseline = np.median(corrfunc)
            baselinedev = mad(corrfunc)
        if self.debug:
            print("baseline, baselinedev:", baseline, baselinedev)

        # then calculate the width of the peak
        if self.peakfittype == "fastquad" or self.peakfittype == "COM":
            peakstart = np.max([1, maxindex - 2])
            peakend = np.min([len(self.corrtimeaxis) - 2, maxindex + 2])
        else:
            # come here for peakfittype of None, quad, gauss, fastgauss
            thegrad = np.gradient(corrfunc).astype(
                "float64"
            )  # the gradient of the correlation function
            if (self.functype == "correlation") or (self.functype == "hybrid"):
                if self.peakfittype == "quad":
                    peakpoints = np.where(
                        corrfunc > maxval_init - 0.05, 1, 0
                    )  # mask for places where correlation exceeds searchfrac*maxval_init
                else:
                    peakpoints = np.where(
                        corrfunc > (baseline + self.searchfrac * (maxval_init - baseline)), 1, 0
                    )  # mask for places where correlation exceeds searchfrac*maxval_init
            else:
                # for mutual information, there is a flattish, nonzero baseline, so we want the difference from that.
                peakpoints = np.where(
                    corrfunc > (baseline + self.searchfrac * (maxval_init - baseline)),
                    1,
                    0,
                )

            peakpoints[0] = 0
            peakpoints[-1] = 0
            peakstart = np.max([1, maxindex - 1])
            peakend = np.min([len(self.corrtimeaxis) - 2, maxindex + 1])
            if self.debug:
                print("initial peakstart, peakend:", peakstart, peakend)
            if self.functype == "mutualinfo":
                while peakpoints[peakend + 1] == 1:
                    peakend += 1
                while peakpoints[peakstart - 1] == 1:
                    peakstart -= 1
            else:
                while (
                    thegrad[peakend + 1] <= 0.0
                    and peakpoints[peakend + 1] == 1
                    and peakend < len(self.corrtimeaxis) - 2
                ):
                    peakend += 1
                while (
                    thegrad[peakstart - 1] >= 0.0
                    and peakpoints[peakstart - 1] == 1
                    and peakstart >= 1
                ):
                    peakstart -= 1
            if self.debug:
                print("final peakstart, peakend:", peakstart, peakend)

            # deal with flat peak top
            while (
                peakend < (len(self.corrtimeaxis) - 3)
                and corrfunc[peakend] == corrfunc[peakend - 1]
            ):
                peakend += 1
            while peakstart > 2 and corrfunc[peakstart] == corrfunc[peakstart + 1]:
                peakstart -= 1
            if self.debug:
                print("peakstart, peakend after flattop correction:", peakstart, peakend)
                print("\n")
                for i in range(peakstart, peakend + 1):
                    print(self.corrtimeaxis[i], corrfunc[i])
                print("\n")
                fig = plt.figure()
                ax = fig.add_subplot(111)
                ax.set_title("Peak sent to fitting routine")
                plt.plot(
                    self.corrtimeaxis[peakstart : peakend + 1],
                    corrfunc[peakstart : peakend + 1],
                    "r",
                )
                plt.show()

            # This is calculated from first principles, but it's always big by a factor or ~1.4.
            #     Which makes me think I dropped a factor if sqrt(2).  So fix that with a final division
            maxsigma_init = np.float64(
                ((peakend - peakstart + 1) * binwidth / (2.0 * np.sqrt(-np.log(self.searchfrac))))
                / np.sqrt(2.0)
            )
            if self.debug:
                print("maxsigma_init:", maxsigma_init)

            # now check the values for errors
            if self.hardlimit:
                rangeextension = 0.0
            else:
                rangeextension = (self.lagmax - self.lagmin) * 0.75
            if not (
                (self.lagmin - rangeextension - binwidth)
                <= maxlag_init
                <= (self.lagmax + rangeextension + binwidth)
            ):
                if maxlag_init <= (self.lagmin - rangeextension - binwidth):
                    failreason |= self.FML_INITLAGLOW
                    maxlag_init = self.lagmin - rangeextension - binwidth
                else:
                    failreason |= self.FML_INITLAGHIGH
                    maxlag_init = self.lagmax + rangeextension + binwidth
                if self.debug:
                    print("bad initial")
            if maxsigma_init > self.absmaxsigma:
                failreason |= self.FML_INITWIDTHHIGH
                maxsigma_init = self.absmaxsigma
                if self.debug:
                    print("bad initial width - too high")
            if peakend - peakstart < 2:
                failreason |= self.FML_INITWIDTHLOW
                maxsigma_init = np.float64(
                    ((2 + 1) * binwidth / (2.0 * np.sqrt(-np.log(self.searchfrac)))) / np.sqrt(2.0)
                )
                if self.debug:
                    print("bad initial width - too low")
            if (self.functype == "correlation") or (self.functype == "hybrid"):
                if not (self.lthreshval <= maxval_init <= self.uthreshval) and self.enforcethresh:
                    failreason |= self.FML_INITAMPLOW
                    if self.debug:
                        print(
                            "bad initial amp:",
                            maxval_init,
                            "is less than",
                            self.lthreshval,
                        )
                if maxval_init < 0.0:
                    failreason |= self.FML_INITAMPLOW
                    maxval_init = 0.0
                    if self.debug:
                        print("bad initial amp:", maxval_init, "is less than 0.0")
                if (maxval_init > 1.0) and self.enforcethresh:
                    failreason |= self.FML_INITAMPHIGH
                    maxval_init = 1.0
                    if self.debug:
                        print("bad initial amp:", maxval_init, "is greater than 1.0")
            else:
                # somewhat different rules for mutual information peaks
                if ((maxval_init - baseline) < self.lthreshval * baselinedev) or (
                    maxval_init < baseline
                ):
                    failreason |= self.FML_INITAMPLOW
                    maxval_init = 0.0
                    if self.debug:
                        print("bad initial amp:", maxval_init, "is less than 0.0")
            if (failreason != self.FML_NOERROR) and self.zerooutbadfit:
                maxval = np.float64(0.0)
                maxlag = np.float64(0.0)
                maxsigma = np.float64(0.0)
            else:
                maxval = np.float64(maxval_init)
                maxlag = np.float64(maxlag_init)
                maxsigma = np.float64(maxsigma_init)

        # refine if necessary
        if self.peakfittype != "None":
            if self.peakfittype == "COM":
                X = self.corrtimeaxis[peakstart : peakend + 1] - baseline
                data = corrfunc[peakstart : peakend + 1]
                maxval = maxval_init
                maxlag = np.sum(X * data) / np.sum(data)
                maxsigma = 10.0
            elif self.peakfittype == "gauss":
                X = self.corrtimeaxis[peakstart : peakend + 1] - baseline
                data = corrfunc[peakstart : peakend + 1]
                # do a least squares fit over the top of the peak
                # p0 = np.array([maxval_init, np.fmod(maxlag_init, lagmod), maxsigma_init], dtype='float64')
                p0 = np.array([maxval_init, maxlag_init, maxsigma_init], dtype="float64")
                if self.debug:
                    print("fit input array:", p0)
                try:
                    plsq, ier = sp.optimize.leastsq(
                        tide_fit.gaussresiduals, p0, args=(data, X), maxfev=5000
                    )
                    if ier not in [1, 2, 3, 4]:  # Check for successful convergence
                        failreason |= self.FML_FITALGOFAIL
                        maxval = np.float64(0.0)
                        maxlag = np.float64(0.0)
                        maxsigma = np.float64(0.0)
                    else:
                        maxval = plsq[0] + baseline
                        maxlag = np.fmod((1.0 * plsq[1]), self.lagmod)
                        maxsigma = plsq[2]
                except:
                    failreason |= self.FML_FITALGOFAIL
                    maxval = np.float64(0.0)
                    maxlag = np.float64(0.0)
                    maxsigma = np.float64(0.0)
                if self.debug:
                    print("fit output array:", [maxval, maxlag, maxsigma])
            elif self.peakfittype == "gausscf":
                X = self.corrtimeaxis[peakstart : peakend + 1] - baseline
                data = corrfunc[peakstart : peakend + 1]
                # do a least squares fit over the top of the peak
                try:
                    plsq, pcov = curve_fit(
                        tide_fit.gaussfunc,
                        X,
                        data,
                        p0=[maxval_init, maxlag_init, maxsigma_init],
                    )
                    maxval = plsq[0] + baseline
                    maxlag = np.fmod((1.0 * plsq[1]), self.lagmod)
                    maxsigma = plsq[2]
                except:
                    failreason |= self.FML_FITALGOFAIL
                    maxval = np.float64(0.0)
                    maxlag = np.float64(0.0)
                    maxsigma = np.float64(0.0)
                if self.debug:
                    print("fit output array:", [maxval, maxlag, maxsigma])
            elif self.peakfittype == "fastgauss":
                X = self.corrtimeaxis[peakstart : peakend + 1] - baseline
                data = corrfunc[peakstart : peakend + 1]
                # do a non-iterative fit over the top of the peak
                # 6/12/2015  This is just broken.  Gives quantized maxima
                maxlag = np.float64(1.0 * np.sum(X * data) / np.sum(data))
                maxsigma = np.float64(
                    np.sqrt(np.abs(np.sum((X - maxlag) ** 2 * data) / np.sum(data)))
                )
                maxval = np.float64(data.max()) + baseline
            elif self.peakfittype == "fastquad":
                maxlag, maxval, maxsigma, ismax, badfit = tide_fit.refinepeak_quad(
                    self.corrtimeaxis, corrfunc, maxindex
                )
            elif self.peakfittype == "quad":
                X = self.corrtimeaxis[peakstart : peakend + 1]
                data = corrfunc[peakstart : peakend + 1]
                try:
                    thecoffs = Polynomial.fit(X, data, 2).convert().coef[::-1]
                    a = thecoffs[0]
                    b = thecoffs[1]
                    c = thecoffs[2]
                    maxlag = -b / (2.0 * a)
                    maxval = a * maxlag * maxlag + b * maxlag + c
                    maxsigma = 1.0 / np.fabs(a)
                    if self.debug:
                        print("poly coffs:", a, b, c)
                        print("maxlag, maxval, maxsigma:", maxlag, maxval, maxsigma)
                except np.exceptions.RankWarning:
                    failreason |= self.FML_FITALGOFAIL
                    maxlag = 0.0
                    maxval = 0.0
                    maxsigma = 0.0
                if self.debug:
                    print("\n")
                    for i in range(len(X)):
                        print(X[i], data[i])
                    print("\n")
                    fig = plt.figure()
                    ax = fig.add_subplot(111)
                    ax.set_title("Peak and fit")
                    plt.plot(X, data, "r")
                    plt.plot(X, c + b * X + a * X * X, "b")
                    plt.show()

            else:
                print("illegal peak refinement type")

            # check for errors in fit
            fitfail = False
            if self.bipolar:
                lowestcorrcoeff = -1.0
            else:
                lowestcorrcoeff = 0.0
            if (self.functype == "correlation") or (self.functype == "hybrid"):
                if maxval < lowestcorrcoeff:
                    failreason |= self.FML_FITAMPLOW
                    maxval = lowestcorrcoeff
                    if self.debug:
                        print("bad fit amp: maxval is lower than lower limit")
                    fitfail = True
                if np.abs(maxval) > 1.0 + self.corrtolerance:
                    failreason |= self.FML_FITAMPHIGH
                    if self.debug:
                        print(
                            "bad fit amp: magnitude of",
                            maxval,
                            "is greater than 1.0",
                        )
                    fitfail = True
                    maxval = 1.0 * np.sign(maxval)
            else:
                # different rules for mutual information peaks
                if ((maxval - baseline) < self.lthreshval * baselinedev) or (maxval < baseline):
                    failreason |= self.FML_FITAMPLOW
                    maxval = 0.0
                    if self.debug:
                        if (maxval - baseline) < self.lthreshval * baselinedev:
                            print(
                                "FITAMPLOW: maxval - baseline:",
                                maxval - baseline,
                                " < lthreshval * baselinedev:",
                                self.lthreshval * baselinedev,
                            )
                        if maxval < baseline:
                            print("FITAMPLOW: maxval < baseline:", maxval, baseline)
                    if self.debug:
                        print("bad fit amp: maxval is lower than lower limit")
            if (self.lagmin > maxlag) or (maxlag > self.lagmax):
                if self.debug:
                    print("bad lag after refinement")
                if self.lagmin > maxlag:
                    failreason |= self.FML_FITLAGLOW
                    maxlag = self.lagmin
                else:
                    failreason |= self.FML_FITLAGHIGH
                    maxlag = self.lagmax
                fitfail = True
            if maxsigma > self.absmaxsigma:
                failreason |= self.FML_FITWIDTHHIGH
                if self.debug:
                    print("bad width after refinement:", maxsigma, ">", self.absmaxsigma)
                maxsigma = self.absmaxsigma
                fitfail = True
            if maxsigma < self.absminsigma:
                failreason |= self.FML_FITWIDTHLOW
                if self.debug:
                    print("bad width after refinement:", maxsigma, "<", self.absminsigma)
                maxsigma = self.absminsigma
                fitfail = True
            if fitfail:
                if self.debug:
                    print("fit fail")
                if self.zerooutbadfit:
                    maxval = np.float64(0.0)
                    maxlag = np.float64(0.0)
                    maxsigma = np.float64(0.0)
                maskval = np.uint16(0)
            # print(maxlag_init, maxlag, maxval_init, maxval, maxsigma_init, maxsigma, maskval, failreason, fitfail)
        else:
            maxval = np.float64(maxval_init)
            maxlag = np.float64(np.fmod(maxlag_init, self.lagmod))
            maxsigma = np.float64(maxsigma_init)
            if failreason != self.FML_NOERROR:
                maskval = np.uint16(0)
            else:
                maskval = np.uint16(1)

        if self.debug or self.displayplots:
            print(
                "init to final: maxval",
                maxval_init,
                maxval,
                ", maxlag:",
                maxlag_init,
                maxlag,
                ", width:",
                maxsigma_init,
                maxsigma,
            )
        if self.displayplots and (self.peakfittype != "None") and (maskval != 0.0):
            fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.set_title("Data and fit")
            hiresx = np.arange(X[0], X[-1], (X[1] - X[0]) / 10.0)
            plt.plot(
                X,
                data,
                "ro",
                hiresx,
                tide_fit.gauss_eval(hiresx, np.array([maxval, maxlag, maxsigma])),
                "b-",
            )
            plt.show()
        return (
            maxindex,
            maxlag,
            flipfac * maxval,
            maxsigma,
            maskval,
            failreason,
            peakstart,
            peakend,
        )


class FrequencyTracker:
    freqs = None
    times = None

    def __init__(
        self,
        lowerlim: float = 0.1,
        upperlim: float = 0.6,
        nperseg: int = 32,
        Q: float = 10.0,
        debug: bool = False,
    ) -> None:
        """
        Initialize the object with spectral analysis parameters.

        Parameters
        ----------
        lowerlim : float, optional
            Lower frequency limit for spectral analysis, default is 0.1
        upperlim : float, optional
            Upper frequency limit for spectral analysis, default is 0.6
        nperseg : int, optional
            Number of samples per segment for spectral analysis, default is 32
        Q : float, optional
            Quality factor for spectral analysis, default is 10.0
        debug : bool, optional
            Debug flag for verbose output, default is False

        Returns
        -------
        None
            This method initializes the object attributes and does not return any value.

        Notes
        -----
        The ``nfft`` attribute is set equal to ``nperseg`` during initialization.

        Examples
        --------
        >>> obj = MyClass(lowerlim=0.2, upperlim=0.8, nperseg=64)
        >>> print(obj.lowerlim)
        0.2
        """
        self.lowerlim = lowerlim
        self.upperlim = upperlim
        self.nperseg = nperseg
        self.Q = Q
        self.debug = debug
        self.nfft = self.nperseg

    def track(self, x: NDArray, fs: float) -> tuple[NDArray, NDArray]:
        """
        Track peak frequencies in a signal using spectrogram analysis and peak fitting.

        This function computes the spectrogram of the input signal, then tracks the
        dominant frequency component over time by fitting peaks in each time segment.
        The result is a tuple of time indices and corresponding peak frequencies.

        Parameters
        ----------
        x : NDArray
            Input signal array to be analyzed.
        fs : float
            Sampling frequency of the input signal in Hz.

        Returns
        -------
        tuple[NDArray, NDArray]
            A tuple containing:
            - times : NDArray
                Time indices corresponding to the tracked peaks (excluding the last time bin).
            - peakfreqs : NDArray
                Array of peak frequencies corresponding to each time segment.
                If no valid peak is found within the specified frequency range,
                the value is set to -1.0.

        Notes
        -----
        - The input signal is padded with zeros at both ends to reduce edge effects.
        - The spectrogram is computed using a Hamming window and no overlap between segments.
        - Peak fitting is performed using a fast quadratic method.
        - Frequencies outside the range defined by `self.lowerlim` and `self.upperlim`
          are marked as invalid (set to -1.0).

        Examples
        --------
        >>> times, peakfreqs = obj.track(signal, fs)
        >>> print(f"Peak frequencies: {peakfreqs}")
        >>> print(f"Time indices: {times}")
        """
        self.freqs, self.times, thespectrogram = sp.signal.spectrogram(
            np.concatenate(
                [np.zeros(int(self.nperseg // 2)), x, np.zeros(int(self.nperseg // 2))],
                axis=0,
            ),
            fs=fs,
            detrend="constant",
            scaling="spectrum",
            nfft=None,
            window=np.hamming(self.nfft),
            noverlap=(self.nperseg - 1),
        )
        lowerliminpts = tide_util.valtoindex(self.freqs, self.lowerlim)
        upperliminpts = tide_util.valtoindex(self.freqs, self.upperlim)

        if self.debug:
            print(self.times.shape, self.freqs.shape, thespectrogram.shape)
            print(self.times)

        # initialize the peak fitter
        thefitter = SimilarityFunctionFitter(
            corrtimeaxis=self.freqs,
            lagmin=self.lowerlim,
            lagmax=self.upperlim,
            absmaxsigma=10.0,
            absminsigma=0.1,
            debug=self.debug,
            peakfittype="fastquad",
            zerooutbadfit=False,
            useguess=False,
        )

        peakfreqs = np.zeros((thespectrogram.shape[1] - 1), dtype=float)
        for i in range(0, thespectrogram.shape[1] - 1):
            (
                maxindex,
                peakfreqs[i],
                maxval,
                maxsigma,
                maskval,
                failreason,
                peakstart,
                peakend,
            ) = thefitter.fit(thespectrogram[:, i])
            if not (lowerliminpts <= maxindex <= upperliminpts):
                peakfreqs[i] = -1.0

        return self.times[:-1], peakfreqs

    def clean(
        self, x: NDArray, fs: float, times: NDArray, peakfreqs: NDArray, numharmonics: int = 2
    ) -> NDArray:
        """
        Apply harmonic cleaning to a signal based on detected peak frequencies and their harmonics.

        This function cleans a signal by applying bandpass filtering to specific time intervals
        centered at given peak frequencies. It supports filtering of harmonics up to a specified
        number and handles edge effects through padding.

        Parameters
        ----------
        x : ndarray
            Input signal to be cleaned.
        fs : float
            Sampling frequency of the signal in Hz.
        times : ndarray
            Array of time indices (in seconds) where cleaning is applied.
        peakfreqs : ndarray
            Array of peak frequencies (in Hz) corresponding to each time index.
        numharmonics : int, optional
            Maximum number of harmonics to filter (default is 2).

        Returns
        -------
        ndarray
            Cleaned signal with harmonics filtered out.

        Notes
        -----
        - The function uses Chebyshev type II filter design for each harmonic.
        - Harmonics are filtered using `scipy.signal.filtfilt` for zero-phase filtering.
        - Edge effects are mitigated by padding the input signal with zeros.

        Examples
        --------
        >>> cleaned_signal = obj.clean(x, fs=1000.0, times=[0.1, 0.5], peakfreqs=[50.0, 100.0])
        """
        nyquistfreq = 0.5 * fs
        y = np.zeros_like(x)
        halfwidth = int(self.nperseg // 2)
        padx = np.concatenate([np.zeros(halfwidth), x, np.zeros(halfwidth)], axis=0)
        pady = np.concatenate([np.zeros(halfwidth), y, np.zeros(halfwidth)], axis=0)
        padweight = np.zeros_like(padx)
        if self.debug:
            print(fs, len(times), len(peakfreqs))
        for i in range(0, len(times)):
            centerindex = int(times[i] * fs)
            xstart = centerindex - halfwidth
            xend = centerindex + halfwidth
            if peakfreqs[i] > 0.0:
                filtsignal = padx[xstart:xend]
                numharmonics = np.min([numharmonics, int((nyquistfreq // peakfreqs[i]) - 1)])
                if self.debug:
                    print("numharmonics:", numharmonics, nyquistfreq // peakfreqs[i])
                for j in range(numharmonics + 1):
                    workingfreq = (j + 1) * peakfreqs[i]
                    if self.debug:
                        print("workingfreq:", workingfreq)
                    ws = [workingfreq * 0.95, workingfreq * 1.05]
                    wp = [workingfreq * 0.9, workingfreq * 1.1]
                    gpass = 1.0
                    gstop = 40.0
                    b, a = sp.signal.iirdesign(wp, ws, gpass, gstop, ftype="cheby2", fs=fs)
                    if self.debug:
                        print(
                            i,
                            j,
                            times[i],
                            centerindex,
                            halfwidth,
                            xstart,
                            xend,
                            xend - xstart,
                            wp,
                            ws,
                            len(a),
                            len(b),
                        )
                    filtsignal = sp.signal.filtfilt(b, a, sp.signal.filtfilt(b, a, filtsignal))
                pady[xstart:xend] += filtsignal
            else:
                pady[xstart:xend] += padx[xstart:xend]
            padweight[xstart:xend] += 1.0
        return (pady / padweight)[halfwidth:-halfwidth]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2021-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
"""Utility functions for rapidtide workflows."""

import logging
import os
from typing import Any, Optional

LGR = logging.getLogger("GENERAL")
TimingLGR = logging.getLogger("TIMING")
MemoryLGR = logging.getLogger("MEMORY")
ErrorLGR = logging.getLogger("ERROR")

starttime = None


class ContextFilter(logging.Filter):
    """
    A filter to allow specific logging handlers to ignore specific loggers.

    We use this to prevent our secondary loggers from printing to the general log file or to
    stdout.
    """

    NAMES = {"TIMING", "MEMORY"}

    def filter(self, record: logging.LogRecord) -> bool:
        if not any([n in record.name for n in self.NAMES]):
            return True
        return False


class TimingFormatter(logging.Formatter):
    """
    Determine whether to filter a log record based on its name.

    This filter checks if any of the names in `self.NAMES` are contained
    within the log record's name. If none of the names are found, the record
    is allowed through (returns True). If any name is found, the record is
    filtered out (returns False).

    Parameters
    ----------
    record : logging.LogRecord
        The log record to be filtered.

    Returns
    -------
    bool
        True if the record should be allowed through (not filtered),
        False if the record should be filtered out.

    Notes
    -----
    The filtering is performed using substring matching. If `self.NAMES` is
    empty, all records will be allowed through.

    Examples
    --------
    >>> class MyFilter:
    ...     NAMES = ['DEBUG', 'INFO']
    ...     def filter(self, record):
    ...         if not any([n in record.name for n in self.NAMES]):
    ...             return True
    ...         return False
    ...
    >>> filter_instance = MyFilter()
    >>> record = logging.LogRecord('DEBUG', 10, 'file.py', 1, 'message', None, None)
    >>> filter_instance.filter(record)
    False
    """

    def format(self, record: logging.LogRecord) -> str:
        """
        Format a log record, extracting message2 and message3 from record.args.

        This method extends the standard logging formatter by checking if the log record's
        args attribute is a dictionary. If so, it extracts "message2" and "message3"
        values from the dictionary and assigns them as attributes to the record for
        use in log formatting.

        Parameters
        ----------
        record : logging.LogRecord
            The log record to be formatted. The record's args attribute is checked
            to determine if message2 and message3 should be extracted.

        Returns
        -------
        str
            The formatted log message string.

        Notes
        -----
        This method modifies the record in-place by adding message2 and message3
        attributes when record.args is a dictionary. If record.args is not a
        dictionary, these attributes are set to None.

        Examples
        --------
        >>> import logging
        >>> formatter = MyFormatter()
        >>> record = logging.LogRecord('name', 1, 'file', 1, 'message', {'message2': 'extra', 'message3': 'info'}, None)
        >>> formatted = formatter.format(record)
        >>> print(formatted)
        'message'
        """
        if isinstance(record.args, dict):
            record.message2 = record.args.get("message2", None)
            record.message3 = record.args.get("message3", None)
        else:
            record.message2 = None
            record.message3 = None
        return super().format(record)


def setup_logger(
    logger_filename: str,
    timing_filename: str,
    memory_filename: Optional[str] = None,
    error_filename: Optional[str] = None,
    isverbose: bool = False,
    debug: bool = False,
) -> None:
    """
    Set up a set of loggers for handling generic, timing, memory, and error logging.

    This function configures multiple loggers with custom handlers and formatters,
    writing output to specified files. It also sets up a custom "VERBOSE" logging level
    and supports optional console output filtering.

    Parameters
    ----------
    logger_filename : str
        Output file for generic logging information.
    timing_filename : str
        Output file for timing-related information.
    memory_filename : str, optional
        Output file for memory usage-related information. If not provided, memory logging is disabled.
    error_filename : str, optional
        Output file for error-related information. If not provided, error logging is disabled.
    isverbose : bool, optional
        Sets the target logging level to VERBOSE (a custom level between INFO and DEBUG).
        Is overridden by ``debug``, if ``debug = True``. Default is False.
    debug : bool, optional
        Sets the target logging level to DEBUG. Default is False.

    Returns
    -------
    None
        This function does not return any value.

    Notes
    -----
    - Existing log files from previous runs are automatically removed.
    - A custom logging level "VERBOSE" (level 15) is added, placed between INFO and DEBUG.
    - Console output is filtered using a `ContextFilter` to avoid duplicate messages.
    - Loggers for timing, memory, and error are configured only if their respective filenames are provided.

    Examples
    --------
    >>> setup_logger(
    ...     logger_filename="app.log",
    ...     timing_filename="timing.log",
    ...     memory_filename="memory.log",
    ...     error_filename="error.log",
    ...     isverbose=True,
    ...     debug=False
    ... )
    """
    # Clean up existing files from previous runs
    for fname in [logger_filename, timing_filename, memory_filename, error_filename]:
        if fname is not None:
            if os.path.isfile(fname):
                LGR.info(f"Removing existing file: {fname}")
                os.remove(fname)

    # Create a new "verbose" logging level
    VERBOSE_LEVEL = 15  # between info and debug
    logging.addLevelName(VERBOSE_LEVEL, "VERBOSE")

    def verbose(self: logging.Logger, message: str, *args: Any, **kwargs: Any) -> None:
        """
        Log a message with VERBOSE_LEVEL if it is enabled.

        This method provides a verbose logging capability that only executes
        when the logger's verbosity level is set to include VERBOSE_LEVEL.

        Parameters
        ----------
        self : logging.Logger
            The logger instance calling this method.
        message : str
            The message to be logged. This message may contain placeholders
            for additional arguments.
        *args : Any
            Additional arguments to be formatted into the message.
        **kwargs : Any
            Additional keyword arguments to be passed to the logging system.

        Returns
        -------
        None
            This method does not return any value.

        Notes
        -----
        This method is a convenience wrapper around the standard logging system
        that only executes when VERBOSE_LEVEL is enabled. It allows for
        conditional verbose logging without requiring explicit checks in the
        calling code.

        Examples
        --------
        >>> import logging
        >>> logger = logging.getLogger(__name__)
        >>> logger.setLevel(logging.DEBUG)
        >>> logger.verbose("Processing item %s", item_name)
        """
        if self.isEnabledFor(VERBOSE_LEVEL):
            self._log(VERBOSE_LEVEL, message, args, **kwargs)

    logging.Logger.verbose = verbose

    # Set logging level for main logger
    if debug:
        logging.root.setLevel(logging.DEBUG)
    elif isverbose:
        logging.root.setLevel(VERBOSE_LEVEL)
    else:
        logging.root.setLevel(logging.INFO)

    # Set up handler for main logger's output file
    log_formatter = logging.Formatter("%(message)s")
    log_handler = logging.FileHandler(logger_filename)
    log_handler.setFormatter(log_formatter)

    # A handler for the console
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(log_formatter)

    # Removing handlers after basicConfig doesn't work, so we use filters
    # for the relevant handlers themselves.
    log_handler.addFilter(ContextFilter())
    stream_handler.addFilter(ContextFilter())

    LGR.addHandler(log_handler)
    LGR.addHandler(stream_handler)
    LGR.propagate = False  # do not print to console, except for messages for StreamHandler

    # A timing logger
    timing_formatter = TimingFormatter(
        "%(asctime)s.%(msecs)03d\t%(message)s\t%(message2)s\t%(message3)s",
        datefmt="%Y%m%dT%H%M%S",
    )
    timing_handler = logging.FileHandler(timing_filename)
    timing_handler.setFormatter(timing_formatter)
    TimingLGR.setLevel(logging.INFO)
    TimingLGR.addHandler(timing_handler)
    TimingLGR.propagate = False  # do not print to console

    # A memory logger
    if memory_filename is not None:
        memory_formatter = logging.Formatter("%(message)s")
        memory_handler = logging.FileHandler(memory_filename)
        memory_handler.setFormatter(memory_formatter)
        MemoryLGR.setLevel(logging.INFO)
        MemoryLGR.addHandler(memory_handler)
        MemoryLGR.propagate = False  # do not print to console

    # An error logger
    if error_filename is not None:
        error_formatter = logging.Formatter("%(message)s")
        error_handler = logging.FileHandler(error_filename)
        error_handler.setFormatter(error_formatter)
        ErrorLGR.setLevel(logging.INFO)
        ErrorLGR.addHandler(error_handler)
        ErrorLGR.propagate = False  # do not print to console



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2024-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
from numpy.typing import NDArray
from scipy.ndimage import binary_erosion

import rapidtide.filter as tide_filt
import rapidtide.stats as tide_stats
from rapidtide.RapidtideDataset import RapidtideDataset


def prepmask(inputmask: NDArray) -> NDArray:
    """
    Apply binary erosion to the input mask.

    Parameters
    ----------
    inputmask : NDArray
        Input binary mask array to be eroded.

    Returns
    -------
    NDArray
        Eroded binary mask array with features reduced by one pixel in all directions.

    Notes
    -----
    This function uses binary erosion to shrink the boundaries of foreground objects
    in the input mask. The erosion operation removes pixels from the boundaries of
    objects, effectively reducing their size by one pixel in all directions.

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.ndimage import binary_erosion
    >>> mask = np.array([[1, 1, 1], [1, 0, 1], [1, 1, 1]])
    >>> eroded = prepmask(mask)
    >>> print(eroded)
    [[0 0 0]
     [0 0 0]
     [0 0 0]]
    """
    erodedmask = binary_erosion(inputmask)
    return erodedmask


def getmasksize(themask: NDArray) -> int:
    """
    Calculate the number of non-zero elements in a mask array.

    This function counts the number of elements in the input array that are greater than zero,
    effectively measuring the size of the active region in a binary mask.

    Parameters
    ----------
    themask : ndarray
        Input array representing a mask, where positive values indicate active regions
        and zero/negative values indicate inactive regions.

    Returns
    -------
    int
        The number of elements in the mask that are greater than zero.

    Notes
    -----
    The function uses `np.where()` to find indices where the mask is greater than zero,
    then `np.ravel()` to flatten the resulting array, and finally `len()` to count the elements.

    Examples
    --------
    >>> import numpy as np
    >>> mask = np.array([[1, 0, 1], [0, 1, 1]])
    >>> getmasksize(mask)
    4

    >>> mask = np.array([0, 0, 0, 0])
    >>> getmasksize(mask)
    0

    >>> mask = np.array([1, 2, 3, 4])
    >>> getmasksize(mask)
    4
    """
    return len(np.ravel(themask[np.where(themask > 0)]))


def checkregressors(
    theregressors: dict[str, Any], numpasses: int, filterlimits: list[float], debug: bool = False
) -> dict[str, float]:
    """
    Calculate and return statistical metrics for the first and last regressors in a sequence.

    This function extracts the first and last regressors from a dictionary of regressors,
    applies spectral filtering based on provided limits, and computes various statistical
    measures such as kurtosis, skewness, and spectral flatness for both regressors.

    Parameters
    ----------
    theregressors : dict[str, Any]
        Dictionary containing regressor objects indexed by pass number (e.g., "pass1", "pass2").
        Each regressor is expected to have attributes like `specaxis`, `specdata`, `kurtosis`,
        `kurtosis_z`, `kurtosis_p`, `skewness`, `skewness_z`, `skewness_p`.
    numpasses : int
        Total number of passes; used to identify the last regressor in the dictionary.
    filterlimits : list[float]
        A list of two floats specifying the lower and upper spectral limits for filtering.
        The function uses these to slice the spectral data.
    debug : bool, optional
        If True, prints debug information including filter limits, indices, and spectral data.
        Default is False.

    Returns
    -------
    dict[str, float]
        A dictionary containing the following keys for both the first and last regressors:
        - `{label}_kurtosis`
        - `{label}_kurtosis_z`
        - `{label}_kurtosis_p`
        - `{label}_skewness`
        - `{label}_skewness_z`
        - `{label}_skewness_p`
        - `{label}_spectralflatness`
        Where `{label}` is either "first" or "last".

    Notes
    -----
    The function uses `numpy.argmax` and `numpy.argmin` to determine the indices of the
    spectral axis that correspond to the provided filter limits. It then slices the
    spectral data using these indices to compute the spectral flatness.

    Examples
    --------
    >>> import numpy as np
    >>> regressors = {
    ...     "pass1": MockRegressor(
    ...         specaxis=np.linspace(0, 10, 100),
    ...         specdata=np.random.rand(100),
    ...         kurtosis=1.0,
    ...         kurtosis_z=0.5,
    ...         kurtosis_p=0.1,
    ...         skewness=0.0,
    ...         skewness_z=0.0,
    ...         skewness_p=0.5
    ...     ),
    ...     "pass2": MockRegressor(
    ...         specaxis=np.linspace(0, 10, 100),
    ...         specdata=np.random.rand(100),
    ...         kurtosis=1.2,
    ...         kurtosis_z=0.6,
    ...         kurtosis_p=0.05,
    ...         skewness=0.1,
    ...         skewness_z=0.1,
    ...         skewness_p=0.4
    ...     )
    ... }
    >>> result = checkregressors(regressors, 2, [2.0, 8.0])
    >>> print(result)
    {'first_kurtosis': 1.0, 'first_kurtosis_z': 0.5, 'first_kurtosis_p': 0.1,
     'first_skewness': 0.0, 'first_skewness_z': 0.0, 'first_skewness_p': 0.5,
     'first_spectralflatness': 0.5, 'last_kurtosis': 1.2, 'last_kurtosis_z': 0.6,
     'last_kurtosis_p': 0.05, 'last_skewness': 0.1, 'last_skewness_z': 0.1,
     'last_skewness_p': 0.4, 'last_spectralflatness': 0.45}
    """
    regressormetrics = {}
    firstregressor = theregressors["pass1"]
    lastregressor = theregressors[f"pass{numpasses}"]
    lowerlimindex = np.argmax(firstregressor.specaxis >= filterlimits[0])
    upperlimindex = np.argmin(firstregressor.specaxis <= filterlimits[1]) + 1
    if debug:
        print(f"{filterlimits=}")
        print(f"{lowerlimindex=}, {upperlimindex=}")
        print(firstregressor.specaxis)
        print(firstregressor.specdata[lowerlimindex:upperlimindex])
    for label, regressor in [["first", firstregressor], ["last", lastregressor]]:
        regressormetrics[f"{label}_kurtosis"] = regressor.kurtosis
        regressormetrics[f"{label}_kurtosis_z"] = regressor.kurtosis_z
        regressormetrics[f"{label}_kurtosis_p"] = regressor.kurtosis_p
        regressormetrics[f"{label}_skewness"] = regressor.skewness
        regressormetrics[f"{label}_skewness_z"] = regressor.skewness_z
        regressormetrics[f"{label}_skewness_p"] = regressor.skewness_p
        regressormetrics[f"{label}_spectralflatness"] = tide_filt.spectralflatness(
            regressor.specdata[lowerlimindex:upperlimindex]
        )
    return regressormetrics


def gethistmetrics(
    themap: NDArray,
    themask: NDArray,
    thedict: dict[str, Any],
    thehistlabel: str | None = None,
    histlen: int = 101,
    rangemin: float = -1.0,
    rangemax: float = 1.0,
    nozero: bool = False,
    savehist: bool = False,
    ignorefirstpoint: bool = False,
    debug: bool = False,
) -> None:
    """
    Compute histogram-based metrics for masked data and store results in a dictionary.

    This function applies a mask to the input data, computes various statistical
    measures including percentiles, moments (skewness, kurtosis), and histogram
    properties, and stores the results in the provided dictionary.

    Parameters
    ----------
    themap : NDArray
        The input data array from which metrics are computed.
    themask : NDArray
        A boolean or numeric mask array used to select valid data points.
    thedict : dict[str, Any]
        Dictionary to store computed metrics. Keys will be updated with statistical
        values such as percentiles, widths, skewness, kurtosis, and histogram properties.
    thehistlabel : str, optional
        Label for the histogram plot if `debug` is enabled. Default is None.
    histlen : int, optional
        Number of bins for the histogram. Default is 101.
    rangemin : float, optional
        Minimum value of the histogram range. Default is -1.0.
    rangemax : float, optional
        Maximum value of the histogram range. Default is 1.0.
    nozero : bool, optional
        If True, exclude zero values from the computation. Default is False.
    savehist : bool, optional
        If True, save histogram bin centers and values to `thedict`. Default is False.
    ignorefirstpoint : bool, optional
        If True, ignore the first point when computing the histogram. Default is False.
    debug : bool, optional
        If True, print debug information and display histogram plot. Default is False.

    Returns
    -------
    None
        Results are stored in the input dictionary `thedict`.

    Notes
    -----
    The function modifies the input dictionary `thedict` in place by adding or updating
    keys with computed statistics. If the mask is empty or no valid data remains after
    filtering, all keys are set to `None` or 0.

    Examples
    --------
    >>> import numpy as np
    >>> data = np.random.rand(100, 100)
    >>> mask = np.ones((100, 100))
    >>> stats_dict = {}
    >>> gethistmetrics(data, mask, stats_dict, debug=True)
    >>> print(stats_dict['pct50'])
    """
    # mask and flatten the data
    maskisempty = False
    if len(np.where(themask > 0)) == 0:
        maskisempty = False
    if debug:
        print("num-nonzero in mask", len(np.where(themask > 0)[0]))
    if not maskisempty:
        dataforhist = np.ravel(themap[np.where(themask > 0.0)])
        if nozero:
            dataforhist = dataforhist[np.where(dataforhist != 0.0)]
            if len(dataforhist) == 0:
                maskisempty = True
            if debug:
                print("num-nonzero in dataforhist", len(dataforhist))
    else:
        maskisempty = True

    if not maskisempty:
        # get percentiles
        (
            thedict["pct02"],
            thedict["pct25"],
            thedict["pct50"],
            thedict["pct75"],
            thedict["pct98"],
        ) = tide_stats.getfracvals(dataforhist, [0.02, 0.25, 0.5, 0.75, 0.98], debug=debug)
        thedict["voxelsincluded"] = len(dataforhist)
        thedict["q1width"] = thedict["pct25"] - thedict["pct02"]
        thedict["q2width"] = thedict["pct50"] - thedict["pct25"]
        thedict["q3width"] = thedict["pct75"] - thedict["pct50"]
        thedict["q4width"] = thedict["pct98"] - thedict["pct75"]
        thedict["mid50width"] = thedict["pct75"] - thedict["pct25"]

        # get moments
        (
            thedict["kurtosis"],
            thedict["kurtosis_z"],
            thedict["kurtosis_p"],
        ) = tide_stats.kurtosisstats(dataforhist)
        (
            thedict["skewness"],
            thedict["skewness_z"],
            thedict["skewness_p"],
        ) = tide_stats.skewnessstats(dataforhist)
        (
            thehist,
            thedict["peakheight"],
            thedict["peakloc"],
            thedict["peakwidth"],
            thedict["centerofmass"],
            thedict["peakpercentile"],
        ) = tide_stats.makehistogram(
            dataforhist,
            histlen,
            refine=False,
            therange=(rangemin, rangemax),
            normalize=True,
            ignorefirstpoint=ignorefirstpoint,
            debug=debug,
        )
        histbincenters = ((thehist[1][1:] + thehist[1][0:-1]) / 2.0).tolist()
        histvalues = (thehist[0][-histlen:]).tolist()
        if savehist:
            thedict["histbincenters"] = histbincenters
            thedict["histvalues"] = histvalues
        if debug:
            fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.set_title(thehistlabel)
            plt.plot(histbincenters, histvalues)
            plt.show()
    else:
        thedict["voxelsincluded"] = 0
        taglist = ["pct02", "pct25", "pct50", "pct75", "pct98"]
        taglist += ["q1width", "q2width", "q3width", "q4width", "mid50width"]
        taglist += ["kurtosis", "kurtosis_z", "kurtosis_p", "skewness", "skewness_z", "skewness_p"]
        taglist += ["peakheight", "peakloc", "peakwidth", "centerofmass", "peakpercentile"]
        if savehist:
            taglist += ["histbincenters", "histvalues"]
        for tag in taglist:
            thedict[tag] = None


def checkmap(
    themap: NDArray,
    themask: NDArray,
    histlen: int = 101,
    rangemin: float = 0.0,
    rangemax: float = 1.0,
    histlabel: str = "similarity metric histogram",
    ignorefirstpoint: bool = False,
    savehist: bool = False,
    debug: bool = False,
) -> dict[str, Any]:
    """
    Compute histogram metrics for a similarity map using a mask.

    This function calculates various statistical metrics from the histogram of
    similarity values in `themap` where `themask` is non-zero. The metrics include
    mean, standard deviation, minimum, maximum, and histogram data.

    Parameters
    ----------
    themap : NDArray
        Array containing similarity values to analyze.
    themask : NDArray
        Binary mask array where non-zero values indicate regions of interest.
    histlen : int, optional
        Number of bins in the histogram (default is 101).
    rangemin : float, optional
        Minimum value for histogram range (default is 0.0).
    rangemax : float, optional
        Maximum value for histogram range (default is 1.0).
    histlabel : str, optional
        Label for the histogram (default is "similarity metric histogram").
    ignorefirstpoint : bool, optional
        Whether to ignore the first point in the histogram calculation (default is False).
    savehist : bool, optional
        Whether to save the histogram data (default is False).
    debug : bool, optional
        Whether to enable debug output (default is False).

    Returns
    -------
    dict[str, Any]
        Dictionary containing histogram metrics including mean, std, min, max,
        and histogram data. The exact keys depend on the implementation of
        `gethistmetrics` function.

    Notes
    -----
    This function serves as a wrapper around `gethistmetrics` and returns the
    computed metrics directly. The histogram is computed only for regions where
    `themask` is non-zero.

    Examples
    --------
    >>> import numpy as np
    >>> map_data = np.random.rand(100, 100)
    >>> mask_data = np.ones((100, 100))
    >>> metrics = checkmap(map_data, mask_data)
    >>> print(metrics.keys())
    dict_keys(['mean', 'std', 'min', 'max', 'hist'])
    """
    themetrics = {}

    gethistmetrics(
        themap,
        themask,
        themetrics,
        thehistlabel=histlabel,
        histlen=histlen,
        rangemin=rangemin,
        rangemax=rangemax,
        nozero=False,
        savehist=savehist,
        ignorefirstpoint=ignorefirstpoint,
        debug=debug,
    )
    return themetrics


def qualitycheck(
    datafileroot: str,
    graymaskspec: str | None = None,
    whitemaskspec: str | None = None,
    anatname: str | None = None,
    geommaskname: str | None = None,
    userise: bool = False,
    usecorrout: bool = False,
    useatlas: bool = False,
    forcetr: bool = False,
    forceoffset: bool = False,
    offsettime: float = 0.0,
    verbose: bool = False,
    debug: bool = False,
) -> dict[str, Any]:
    """
    Perform quality checks on a dataset by analyzing masks, regressors, and map statistics.

    This function loads a dataset using `RapidtideDataset` and performs a series of quality
    assessments on various overlays (e.g., lag times, strengths, MTT) and regressors. It
    computes statistics for different masks and map regions, including histogram data and
    relative sizes. Optional gray and white matter masks can be used to isolate analysis
    within those regions.

    Parameters
    ----------
    datafileroot : str
        Root name of the data files to be processed.
    graymaskspec : str, optional
        Path to the gray matter mask specification file.
    whitemaskspec : str, optional
        Path to the white matter mask specification file.
    anatname : str, optional
        Name of the anatomical image to use.
    geommaskname : str, optional
        Name of the geometric mask to use.
    userise : bool, default=False
        Whether to use RISE (reconstruction of instantaneous signal estimates).
    usecorrout : bool, default=False
        Whether to use corrected output.
    useatlas : bool, default=False
        Whether to use atlas-based registration.
    forcetr : bool, default=False
        Force TR (repetition time) to be set.
    forceoffset : bool, default=False
        Force offset to be set.
    offsettime : float, default=0.0
        Time offset to apply.
    verbose : bool, default=False
        Enable verbose output.
    debug : bool, default=False
        Enable debug output.

    Returns
    -------
    dict[str, Any]
        A dictionary containing quality check results, including:
        - ``passes``: Number of passes in the dataset.
        - ``filterlimits``: Regressor filter limits.
        - ``simcalclimits``: Regressor similarity calculation limits.
        - ``mask``: Dictionary of mask-related statistics.
        - ``regressor``: Regressor quality check results.
        - ``lag``, ``laggrad``, ``strength``, ``MTT``: Statistics for respective maps.
        - Optional gray/white matter-specific results if masks are provided.

    Notes
    -----
    This function relies on several helper functions and classes such as `RapidtideDataset`,
    `prepmask`, `checkregressors`, and `checkmap`. It uses the `numpy` library for array
    operations and `matplotlib` for histogram plotting (when enabled).

    Examples
    --------
    >>> output = qualitycheck("sub-01", graymaskspec="gray_mask.nii.gz")
    >>> print(output["passes"])
    3
    >>> print(output["mask"]["refinemaskrelsize"])
    0.75
    """
    # read in the dataset
    thedataset = RapidtideDataset(
        "main",
        datafileroot + "_",
        anatname=anatname,
        geommaskname=geommaskname,
        graymaskspec=graymaskspec,
        whitemaskspec=whitemaskspec,
        userise=userise,
        usecorrout=usecorrout,
        useatlas=useatlas,
        forcetr=forcetr,
        forceoffset=forceoffset,
        offsettime=offsettime,
        verbose=verbose,
        init_LUT=False,
    )

    if debug:
        print(f"qualitycheck started on {datafileroot}")
    outputdict = {}
    if graymaskspec is not None:
        dograyonly = True
        thegraymask = (thedataset.overlays["graymask"]).data
    else:
        dograyonly = False
    if whitemaskspec is not None:
        dowhiteonly = True
        thewhitemask = (thedataset.overlays["whitemask"]).data
    else:
        dowhiteonly = False

    # put in some basic information
    outputdict["passes"] = thedataset.numberofpasses
    outputdict["filterlimits"] = thedataset.regressorfilterlimits
    outputdict["simcalclimits"] = thedataset.regressorsimcalclimits

    # process the masks
    outputdict["mask"] = {}
    thelagmask = (thedataset.overlays["lagmask"]).data
    theerodedmask = prepmask(thelagmask)
    outputdict["mask"]["lagmaskvoxels"] = len(np.ravel(thelagmask[np.where(thelagmask > 0)]))
    for maskname in [
        "refinemask",
        "meanmask",
        "preselectmask",
        "p_lt_0p050_mask",
        "p_lt_0p010_mask",
        "p_lt_0p005_mask",
        "p_lt_0p001_mask",
        "desc-plt0p001_mask",
    ]:
        try:
            thismask = (thedataset.overlays[maskname]).data
        except KeyError:
            print(f"{maskname} not found in dataset")
        else:
            outname = maskname.replace("_mask", "").replace("mask", "")
            outputdict["mask"][f"{outname}relsize"] = getmasksize(thismask) / (
                1.0 * outputdict["mask"]["lagmaskvoxels"]
            )

    # process the regressors
    theregressors = thedataset.regressors
    outputdict["regressor"] = checkregressors(
        theregressors, outputdict["passes"], outputdict["filterlimits"], debug=debug
    )

    # process the lag map
    thelags = thedataset.overlays["lagtimes"]
    thelags.setFuncMask(thelagmask)
    thelags.updateStats()
    if debug:
        thelags.summarize()
    outputdict["lag"] = checkmap(
        thelags.data,
        thelagmask,
        rangemin=-5.0,
        rangemax=10.0,
        histlabel="lag histogram",
        debug=debug,
    )

    # get the gradient of the lag map
    thegradient = np.gradient(thelags.data)
    thegradientamp = np.sqrt(
        np.square(thegradient[0] / thelags.xsize)
        + np.square(thegradient[1] / thelags.ysize)
        + np.square(thegradient[2] / thelags.zsize)
    )
    outputdict["laggrad"] = checkmap(
        thegradientamp,
        theerodedmask,
        rangemin=0.0,
        rangemax=3.0,
        histlabel="lag gradient amplitude histogram",
        ignorefirstpoint=True,
        debug=debug,
    )

    # process the strength map
    thestrengths = thedataset.overlays["lagstrengths"]
    thestrengths.setFuncMask(thelagmask)
    thestrengths.updateStats()
    if debug:
        thestrengths.summarize()
    outputdict["strength"] = checkmap(
        thestrengths.data,
        thelagmask,
        rangemin=0.0,
        rangemax=1.0,
        histlabel="similarity metric histogram",
        debug=debug,
    )

    # process the MTT map
    theMTTs = thedataset.overlays["MTT"]
    theMTTs.setFuncMask(thelagmask)
    theMTTs.updateStats()
    if debug:
        theMTTs.summarize()
    outputdict["MTT"] = checkmap(
        theMTTs.data,
        thelagmask,
        histlabel="MTT histogram",
        rangemin=0.0,
        rangemax=10.0,
        debug=debug,
    )

    if dograyonly:
        outputdict["grayonly-lag"] = checkmap(
            thelags.data,
            thelagmask * thegraymask,
            rangemin=-5.0,
            rangemax=10.0,
            histlabel="lag histogram - gray only",
            debug=debug,
        )
        outputdict["grayonly-laggrad"] = checkmap(
            thegradientamp,
            theerodedmask * thegraymask,
            rangemin=0.0,
            rangemax=3.0,
            histlabel="lag gradient amplitude histogram - gray only",
            debug=debug,
        )
        outputdict["grayonly-strength"] = checkmap(
            thestrengths.data,
            thelagmask * thegraymask,
            rangemin=0.0,
            rangemax=1.0,
            histlabel="similarity metric histogram - gray only",
            debug=debug,
        )
    if dowhiteonly:
        outputdict["whiteonly-lag"] = checkmap(
            thelags.data,
            thelagmask * thewhitemask,
            rangemin=-5.0,
            rangemax=10.0,
            histlabel="lag histogram - white only",
            debug=debug,
        )
        outputdict["whiteonly-laggrad"] = checkmap(
            thegradientamp,
            theerodedmask * thewhitemask,
            rangemin=0.0,
            rangemax=3.0,
            histlabel="lag gradient amplitude histogram - white only",
            debug=debug,
        )
        outputdict["whiteonly-strength"] = checkmap(
            thestrengths.data,
            thelagmask * thewhitemask,
            rangemin=0.0,
            rangemax=1.0,
            histlabel="similarity metric histogram - white only",
            debug=debug,
        )

    return outputdict



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import argparse
import copy
from typing import Any

import matplotlib.pyplot as plt
import numpy as np

import rapidtide.io as tide_io
import rapidtide.maskutil as tide_mask
import rapidtide.stats as tide_stats
import rapidtide.workflows.parser_funcs as pf

DEFAULT_LAGMIN = -10
DEFAULT_LAGMAX = 20
DEFAULT_PEAKTHRESH = 0.33
DEFAULT_HISTBINS = 151


def _get_parser() -> Any:
    """
    Argument parser for adjust offset.

    This function constructs and returns an `argparse.ArgumentParser` object configured
    for parsing command-line arguments used by the `adjustoffset` tool. It defines
    various options for adjusting the offset of a rapidtide delay map, including
    masking, histogram-based peak detection, and output control.

    Returns
    -------
    argparse.ArgumentParser
        Configured argument parser for the adjustoffset tool.

    Notes
    -----
    The parser includes support for:
    - Input and output file specifications
    - Masking options (include, exclude, extra)
    - Histogram-based offset estimation
    - Search range limiting
    - Debugging and display options

    Examples
    --------
    >>> parser = _get_parser()
    >>> args = parser.parse_args(['input.nii', 'output_root'])
    >>> print(args.inputmap)
    'input.nii'
    """
    parser = argparse.ArgumentParser(
        prog="adjustoffset",
        description="Adjust the offset of a rapidtide delay map.",
        allow_abbrev=False,
    )

    # Required arguments
    parser.add_argument(
        "inputmap",
        type=lambda x: pf.is_valid_file(parser, x),
        help="The name of the rapidtide maxtime map.",
    )
    parser.add_argument("outputroot", help="The root name for the output files.")

    maskopts = parser.add_argument_group("Masking options")
    maskopts.add_argument(
        "--includemask",
        dest="includespec",
        metavar="MASK[:VALSPEC]",
        help=(
            "Only use voxels that are also in file MASK in calculating the offset values "
            "(if VALSPEC is given, only voxels "
            "with integral values listed in VALSPEC are used). "
        ),
        default=None,
    )
    maskopts.add_argument(
        "--excludemask",
        dest="excludespec",
        metavar="MASK[:VALSPEC]",
        help=(
            "Do not use voxels that are also in file MASK in calculating the offset values "
            "(if VALSPEC is given, voxels "
            "with integral values listed in VALSPEC are excluded). "
        ),
        default=None,
    )
    maskopts.add_argument(
        "--extramask",
        dest="extramaskname",
        metavar="MASK",
        type=lambda x: pf.is_valid_file(parser, x),
        help=(
            "Additional mask to apply to select voxels for adjustment.  Zero voxels in this mask will be excluded.  "
            "If not specified, the corrfit_mask will be used."
        ),
        default=None,
    )

    parser.add_argument(
        "--searchrange",
        dest="searchrange",
        action=pf.IndicateSpecifiedAction,
        nargs=2,
        type=float,
        metavar=("LAGMIN", "LAGMAX"),
        help=(
            "Limit fit to a range of lags from LAGMIN to "
            f"LAGMAX.  Default is {DEFAULT_LAGMIN} to {DEFAULT_LAGMAX} seconds. "
        ),
        default=(DEFAULT_LAGMIN, DEFAULT_LAGMAX),
    )
    parser.add_argument(
        "--histbins",
        metavar="BINS",
        help=f"Number of bins in the entropy histogram (default is {DEFAULT_HISTBINS}).",
        type=int,
        default=DEFAULT_HISTBINS,
    )
    parser.add_argument(
        "--histonly",
        dest="histonly",
        action="store_true",
        help=("Only calculate offset histograms - do not perform adjustments."),
        default=False,
    )
    parser.add_argument(
        "--display",
        dest="display",
        action="store_true",
        help=("Show the delay histogram."),
        default=False,
    )
    parser.add_argument(
        "--pickleft",
        dest="pickleft",
        action="store_true",
        help=("Choose the leftmost peak of the histogram that exceeds the threshold."),
        default=False,
    )
    parser.add_argument(
        "--pickleftthresh",
        metavar="THRESH",
        help=f"Fraction of the maximum height that can be considered a peak.  Default is {DEFAULT_PEAKTHRESH}",
        type=float,
        default=DEFAULT_PEAKTHRESH,
    )
    parser.add_argument(
        "--setoffset",
        metavar="OFFSET",
        help=f"Directly set the offset value to OFFSET.  Overrides histogram.",
        type=float,
        default=None,
    )
    parser.add_argument(
        "--norefine",
        dest="refine",
        action="store_false",
        help=("Do not fit the histogram peak."),
        default=True,
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        help=("Output debugging information."),
        default=False,
    )

    return parser


def adjustoffset(args: Any) -> None:
    """
    Adjust the offset of a NIfTI map based on histogram analysis and optional masking.

    This function reads a NIfTI map file, applies optional inclusion and exclusion masks,
    and computes a peak location from the histogram of valid voxels. The computed offset
    is then added to the map values, unless a fixed offset is specified via `args.setoffset`.

    Parameters
    ----------
    args : Any
        An object containing the following attributes:
        - `inputmap` : str
            Path to the input NIfTI map file.
        - `debug` : bool
            If True, prints debug information.
        - `includespec` : str, optional
            Specification for including voxels in the analysis.
        - `excludespec` : str, optional
            Specification for excluding voxels from the analysis.
        - `extramaskname` : str, optional
            Path to an additional mask file.
        - `histbins` : int
            Number of histogram bins to use.
        - `searchrange` : tuple of float, optional
            Range of values to consider for histogram analysis.
        - `refine` : bool
            Whether to refine the peak detection.
        - `pickleft` : bool
            Whether to pick the leftmost peak.
        - `pickleftthresh` : float, optional
            Threshold for leftmost peak picking.
        - `display` : bool
            Whether to display the histogram.
        - `histonly` : bool
            If True, only compute and display the histogram, do not adjust the map.
        - `setoffset` : float, optional
            Fixed offset value to apply to the map.
        - `outputroot` : str
            Root name for output files.

    Returns
    -------
    None
        This function does not return a value but saves two NIfTI files:
        - `<outputroot>_maskmap.nii.gz`: The generated mask map.
        - `<outputroot>_adjustedmaxtime.nii.gz`: The adjusted map with offset applied.

    Notes
    -----
    - The function uses `tide_io.readfromnifti` to read the input map and `tide_io.savetonifti` to save outputs.
    - Masks are generated using `tide_mask.getmaskset` based on inclusion/exclusion specifications.
    - Histogram analysis is performed using `tide_stats.gethistprops` and `tide_stats.makehistogram`.
    - If `args.setoffset` is provided, it overrides the computed peak location as the offset.

    Examples
    --------
    >>> class Args:
    ...     inputmap = "input_map.nii.gz"
    ...     debug = True
    ...     includespec = "brain"
    ...     excludespec = None
    ...     extramaskname = None
    ...     histbins = 100
    ...     searchrange = (0.0, 10.0)
    ...     refine = True
    ...     pickleft = False
    ...     pickleftthresh = 0.5
    ...     display = False
    ...     histonly = False
    ...     setoffset = None
    ...     outputroot = "output"
    ...
    >>> args = Args()
    >>> adjustoffset(args)
    """
    if args.debug:
        print(f"reading map file {args.inputmap}")
    (
        themap,
        themap_data,
        themap_hdr,
        themapdims,
        thetemplatesizes,
    ) = tide_io.readfromnifti(args.inputmap)
    nx, ny, nz, nummaps = tide_io.parseniftidims(themapdims)

    # process masks
    if args.includespec is not None:
        (
            includename,
            includevals,
        ) = tide_io.processnamespec(
            args.includespec, "Including voxels where ", "in offset calculation."
        )
    else:
        includename = None
        includevals = None
    if args.excludespec is not None:
        (
            excludename,
            excludevals,
        ) = tide_io.processnamespec(
            args.excludespec, "Excluding voxels where ", "from offset calculation."
        )
    else:
        excludename = None
        excludevals = None

    if args.extramaskname is None:
        args.extramaskname = (args.inputmap).replace("maxtime_map", "corrfit_mask")

    numspatiallocs = int(nx) * int(ny) * int(nz)
    includemask, excludemask, extramask = tide_mask.getmaskset(
        "anatomic",
        includename,
        includevals,
        excludename,
        excludevals,
        themap_hdr,
        numspatiallocs,
        extramask=args.extramaskname,
    )

    theflatmap = themap_data.reshape((numspatiallocs))
    theflatmask = np.ones_like(theflatmap)
    if includemask is not None:
        theflatmask = theflatmask * includemask.reshape((numspatiallocs))
    if excludemask is not None:
        theflatmask = theflatmask * (1 - excludemask.reshape((numspatiallocs)))
    if extramask is not None:
        theflatextramask = extramask.reshape((numspatiallocs))
        theflatmask = theflatmask * theflatextramask

    # generate the mask
    themask_data = theflatmask.reshape((nx, ny, nz))
    maskmap = themask_data

    peakloc, peakheight, peakwidth = tide_stats.gethistprops(
        theflatmap[np.where(theflatmask > 0.0)],
        args.histbins,
        therange=args.searchrange,
        refine=args.refine,
        pickleft=args.pickleft,
        peakthresh=args.pickleftthresh,
    )
    print(f"{peakloc=}, {peakheight=}, {peakwidth=}")

    (
        thehist,
        peakheight2,
        peakloc2,
        peakwidth2,
        centerofmass2,
        peakpercentile2,
    ) = tide_stats.makehistogram(
        theflatmap[np.where(theflatmask > 0.0)],
        args.histbins,
        binsize=None,
        therange=args.searchrange,
        refine=args.refine,
        normalize=False,
    )
    # print(f"{peakloc=}, {peakheight=}, {peakwidth=}, {centerofmass=}")

    endtrim = 1
    if args.display:
        thestore = np.zeros((2, len(thehist[0])), dtype="float64")
        thestore[0, :] = (thehist[1][1:] + thehist[1][0:-1]) / 2.0
        thestore[1, :] = thehist[0][-args.histbins :]
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_title("histogram")
        plt.plot(thestore[0, : (-1 - endtrim)], thestore[1, : (-1 - endtrim)])
        plt.show()

    # save the maskmap as nifti
    themaskmaphdr = copy.deepcopy(themap_hdr)
    themaskmaphdr["dim"][0] = 3
    themaskmaphdr["dim"][4] = 1
    tide_io.savetonifti(maskmap, themaskmaphdr, args.outputroot + "_maskmap")

    if not args.histonly:
        # actually change the offset
        if args.setoffset is not None:
            offsetvalue = args.setoffset
        else:
            offsetvalue = peakloc
        theflatmap[np.where(theflatmask > 0.0)] += offsetvalue
        tide_io.savetonifti(
            theflatmap.reshape((nx, ny, nz)), themap_hdr, args.outputroot + "_adjustedmaxtime"
        )
